from __future__ import annotations

from .clean import clean_data
from .configs import (
    ClassifierEnsembleConfig,
    EnsembleConfig,
    PreprocessorConfig,
    RegressorEnsembleConfig,
)
from .ensemble import (
    generate_classification_ensemble_configs,
    generate_regression_ensemble_configs,
)
from .pipeline_interface import (
    PreprocessingPipeline,
)
from .presets import (
    default_classifier_preprocessor_configs,
    default_regressor_preprocessor_configs,
    v2_5_classifier_preprocessor_configs,
    v2_5_regressor_preprocessor_configs,
    v2_classifier_preprocessor_configs,
    v2_regressor_preprocessor_configs,
)
from .transform import fit_preprocessing

__all__ = [
    "ClassifierEnsembleConfig",
    "EnsembleConfig",
    "PreprocessingPipeline",
    "PreprocessorConfig",
    "RegressorEnsembleConfig",
    "clean_data",
    "default_classifier_preprocessor_configs",
    "default_regressor_preprocessor_configs",
    "fit_preprocessing",
    "generate_classification_ensemble_configs",
    "generate_regression_ensemble_configs",
    "v2_5_classifier_preprocessor_configs",
    "v2_5_regressor_preprocessor_configs",
    "v2_classifier_preprocessor_configs",
    "v2_regressor_preprocessor_configs",
]



"""Tests that cover both the classification and regression interfaces."""

from __future__ import annotations

import platform

import numpy as np
import pytest
import torch

from tabpfn import TabPFNClassifier, TabPFNRegressor
from tabpfn.constants import ModelVersion
from tests.utils import get_pytest_devices

devices = get_pytest_devices()

device_combinations = [
    (devices[0], devices[-1]),
    # Use different cpu indicies because the same device can't appear twice. This seems
    # to work, even if there's only one cpu.
    ("auto", ["cpu:0", "cpu:1"]),
]


@pytest.mark.parametrize(("device_1", "device_2"), device_combinations)
@pytest.mark.parametrize("estimator_class", [TabPFNRegressor, TabPFNClassifier])
@pytest.mark.parametrize(
    "fit_mode", ["fit_preprocessors", "fit_with_cache", "low_memory"]
)
def test__to__before_fit__does_not_crash(
    estimator_class: type[TabPFNClassifier] | type[TabPFNRegressor],
    fit_mode: str,
    device_1: str,
    device_2: str,
) -> None:
    estimator = estimator_class(fit_mode=fit_mode, device=device_1, n_estimators=2)
    X_train, X_test, y_train = _get_tiny_dataset(estimator)
    estimator.to(device_2)
    estimator.fit(X_train, y_train)
    estimator.predict(X_test)


@pytest.mark.parametrize(("device_1", "device_2"), device_combinations)
@pytest.mark.parametrize("estimator_class", [TabPFNRegressor, TabPFNClassifier])
@pytest.mark.parametrize("fit_mode", ["fit_preprocessors", "low_memory"])
def test__to__between_fit_and_predict__does_not_crash(
    estimator_class: type[TabPFNClassifier] | type[TabPFNRegressor],
    fit_mode: str,
    device_1: str,
    device_2: str,
) -> None:
    estimator = estimator_class(fit_mode=fit_mode, device=device_1, n_estimators=2)
    X_train, X_test, y_train = _get_tiny_dataset(estimator)
    estimator.fit(X_train, y_train)
    estimator.to(device_2)
    estimator.predict(X_test)


@pytest.mark.parametrize(("device_1", "device_2"), device_combinations)
@pytest.mark.parametrize("estimator_class", [TabPFNRegressor, TabPFNClassifier])
@pytest.mark.parametrize("fit_mode", ["fit_preprocessors", "low_memory"])
def test__to__between_fits__outputs_equal(
    estimator_class: type[TabPFNClassifier] | type[TabPFNRegressor],
    fit_mode: str,
    device_1: str,
    device_2: str,
) -> None:
    estimator = estimator_class(
        fit_mode=fit_mode,
        device=device_1,
        n_estimators=2,
        # MPS doesn't support float64, so use a lower precision in that case.
        inference_precision="auto" if platform.system() == "Darwin" else torch.float64,
    )
    X_train, X_test, y_train = _get_tiny_dataset(estimator)
    estimator.fit(X_train, y_train)
    prediction_1 = estimator.predict(X_test)
    estimator.to(device_2)
    estimator.fit(X_train, y_train)
    prediction_2 = estimator.predict(X_test)

    if isinstance(estimator, TabPFNRegressor) and "mps" in devices:
        # Skip only at this point to check that calling .fit() and .to() in this order
        # doesn't cause a crash.
        pytest.skip("MPS yields different predictions.")

    np.testing.assert_array_almost_equal(
        prediction_1,
        prediction_2,
        # Use a slightly relaxed comparison as comparing between devices.
        decimal=4,
    )


@pytest.mark.parametrize("estimator_class", [TabPFNRegressor, TabPFNClassifier])
def test__to__fit_with_cache_and_after_first_fit__raises_error(
    estimator_class: type[TabPFNClassifier] | type[TabPFNRegressor],
) -> None:
    estimator = estimator_class(fit_mode="fit_with_cache", n_estimators=2)
    X_train, X_test, y_train = _get_tiny_dataset(estimator)

    estimator.fit(X_train, y_train)
    with pytest.raises(NotImplementedError):
        estimator.to("cpu")

    estimator.predict(X_test)
    with pytest.raises(NotImplementedError):
        estimator.to("cpu")


@pytest.mark.parametrize("estimator_class", [TabPFNRegressor, TabPFNClassifier])
@pytest.mark.parametrize("fit_mode", ["fit_preprocessors", "low_memory"])
def test__to__after_fit__no_tensors_left_on_old_device(
    estimator_class: type[TabPFNClassifier] | type[TabPFNRegressor],
    fit_mode: str,
) -> None:
    alt_device = "cuda" if "cuda" in devices else "mps" if "mps" in devices else None
    if alt_device is None:
        pytest.skip("Test can only run when two devices are available.")

    estimator = estimator_class(fit_mode=fit_mode, device=alt_device, n_estimators=2)
    X_train, _X_test, y_train = _get_tiny_dataset(estimator)
    estimator.fit(X_train, y_train)
    estimator.to("cpu")

    tensors_not_on_cpu = _find_tensors_not_on_cpu(estimator)
    assert not tensors_not_on_cpu, f"Found tensors not on cpu: {tensors_not_on_cpu}"


@pytest.mark.parametrize("estimator_class", [TabPFNRegressor, TabPFNClassifier])
@pytest.mark.parametrize("fit_mode", ["fit_preprocessors", "low_memory"])
@pytest.mark.parametrize("model_version", [ModelVersion.V2, ModelVersion.V2_5])
def test__to__after_fit_and_predict__no_tensors_left_on_old_device(
    estimator_class: type[TabPFNClassifier] | type[TabPFNRegressor],
    fit_mode: str,
    model_version: ModelVersion,
) -> None:
    alt_device = "cuda" if "cuda" in devices else "mps" if "mps" in devices else None
    if alt_device is None:
        pytest.skip("Test can only run when two devices are available.")

    estimator = estimator_class.create_default_for_version(
        model_version, fit_mode=fit_mode, device=alt_device, n_estimators=2
    )
    X_train, X_test, y_train = _get_tiny_dataset(estimator)
    estimator.fit(X_train, y_train)
    estimator.predict(X_test)
    estimator.to("cpu")

    tensors_not_on_cpu = _find_tensors_not_on_cpu(estimator)
    assert not tensors_not_on_cpu, f"Found tensors not on cpu: {tensors_not_on_cpu}"


def _find_tensors_not_on_cpu(
    estimator: TabPFNClassifier | TabPFNRegressor,
    path: str = "root",
    visited: set[int] | None = None,
) -> list[str]:
    if visited is None:
        visited = set()

    obj_id = id(estimator)
    if obj_id in visited:
        return []
    visited.add(obj_id)

    results: list[str] = []

    if isinstance(estimator, torch.Tensor):
        if estimator.device.type != "cpu":
            results.append(f"{path} (device={estimator.device})")
        return results

    if hasattr(estimator, "__dict__"):
        for attr_name, attr_value in estimator.__dict__.items():
            results.extend(
                _find_tensors_not_on_cpu(attr_value, f"{path}.{attr_name}", visited)
            )

    if isinstance(estimator, dict):
        for key, value in estimator.items():
            results.extend(_find_tensors_not_on_cpu(value, f"{path}[{key!r}]", visited))
    elif isinstance(estimator, (list, tuple)):
        for i, item in enumerate(estimator):
            results.extend(_find_tensors_not_on_cpu(item, f"{path}[{i}]", visited))

    return results


def _get_tiny_dataset(
    estimator: TabPFNClassifier | TabPFNRegressor,
) -> tuple[np.ndarray, np.ndarray, np.ndarray]:
    n_train = 4
    n_test = 2
    generator = np.random.default_rng(seed=0)
    X = generator.normal(loc=0, scale=1, size=(n_train + n_test, 3))
    if isinstance(estimator, TabPFNClassifier):
        y_train = generator.integers(0, 1, size=n_train)
    elif isinstance(estimator, TabPFNRegressor):
        y_train = generator.normal(loc=0, scale=1, size=n_train)
    return X[:n_train], X[n_train:], y_train



"""Encoder step to remove empty (constant) features."""

#  Copyright (c) Prior Labs GmbH 2025.

from __future__ import annotations

from typing import Any
from typing_extensions import override

import torch

from tabpfn.architectures.encoders import TorchPreprocessingStep

from ._ops import select_features


class RemoveEmptyFeaturesEncoderStep(TorchPreprocessingStep):
    """Encoder step to remove empty (constant) features."""

    def __init__(
        self,
        in_keys: tuple[str, ...] = ("main",),
        out_keys: tuple[str, ...] = ("main",),
    ):
        """Initialize the RemoveEmptyFeaturesEncoderStep.

        Args:
            in_keys: The keys of the input tensors.
            out_keys: The keys to assign the output tensors to.
        """
        assert len(in_keys) == len(out_keys) == 1, (
            f"{self.__class__.__name__} expects a single input and output key."
        )
        super().__init__(in_keys, out_keys)
        self.register_buffer("column_selection_mask", None, persistent=False)

    @override
    def _fit(
        self,
        state: dict[str, torch.Tensor],
        **kwargs: Any,
    ) -> None:
        """Compute the non-empty feature selection mask on the training set.

        Args:
            state: The state dict containing the input tensor under `in_keys[0]`.
            **kwargs: Additional keyword arguments (unused).
        """
        del kwargs
        x = state[self.in_keys[0]]
        self.column_selection_mask = (x[1:] == x[0]).sum(0) != (x.shape[0] - 1)

    @override
    def _transform(
        self,
        state: dict[str, torch.Tensor],
        **kwargs: Any,
    ) -> dict[str, torch.Tensor]:
        """Remove empty features from the input tensor.

        Args:
            state: The state dict containing the input tensor under `in_keys[0]`.
            **kwargs: Additional keyword arguments (unused).

        Returns:
            A dict mapping `out_keys[0]` to the transformed tensor with empty features
            removed.
        """
        del kwargs
        x = state[self.in_keys[0]]

        orig_last_dim = x.shape[-1]
        # Ensure that the mask is a bool, because the buffer may get converted to a
        # a float if .to() is called on the containing module.
        x = select_features(x, self.column_selection_mask.type(torch.bool))

        potential_padding = -x.shape[-1] % orig_last_dim
        x = torch.nn.functional.pad(x, pad=(0, potential_padding), value=0)

        return {self.out_keys[0]: x}



#  Copyright (c) Prior Labs GmbH 2025.

"""Interfaces for encoders."""

from __future__ import annotations

import abc
from typing import Any
from typing_extensions import override

import torch
from torch import nn


# Note that inheriting from nn.Sequential is not strictly necessary, because
# we don't want to include learnable parameters in this pipeline.
# Because previous TabPFN checkpoints used nn.Sequential and contain
# keys like `encoder.5.layer.weight` in the state_dict, we keep the inheritance
# for now.
class TorchPreprocessingPipeline(torch.nn.Sequential):
    """Container for a sequence of GPU preprocessing steps.

    These transformations can alter feature sets or values or add
    additional "encoding" like nan indicators to the input.
    These steps must be non-learnable (i.e., no trainable weights or model
    parameters) and strictly used for data preparation.
    """

    def __init__(
        self,
        steps: list[TorchPreprocessingStep],
        # output_key is set for backwards compatibility
        # with encoders that still have the projections in this pipeline.
        output_key: str | None = None,
    ):
        super().__init__(*steps)
        self.output_key = output_key

        # If we are in backwards compatibility mode, we expect the last step to
        # be a LinearInputEncoderStep or MLPInputEncoderStep that combine
        # the entries of the state dict parsed through the encoder.
        if self.output_key is not None:
            assert any(
                s.__class__.__name__
                # Avoid circular import by checking the class name instead of importing
                in ("LinearInputEncoderStep", "MLPInputEncoderStep")
                for s in self
            ), "output_key is only supported for LinearInputEncoderStep"

    # For now, we disable compilation for the preprocessing pipeline because
    # there are multiple data-dependent control flows in the steps that break the graph.
    @torch.compiler.disable
    @override
    def forward(
        self,
        input: dict[str, torch.Tensor] | torch.Tensor,
        single_eval_pos: int | None = None,
        **kwargs: Any,
    ) -> dict[str, torch.Tensor] | torch.Tensor:
        """Fit and transform the preprocessing pipeline on input data.

        Args:
            input: The input tensor or dictionary of tensors in the case of
                multiple modalities. If a tensor is provided, it is wrapped in a
                dictionary with the key "main". The tensor must have the shape
                [R, B * G, F] with
                R = number of rows (train + test),
                B = batch size
                G = number of feature groups
                F = number of features per group
            single_eval_pos: The position to use to split train and test data.
            **kwargs: Additional keyword arguments passed to the encoder step.

        Returns:
            A dictionary of tensors with all keys created in the pipeline or
            (for backwards compatibility only) the output tensor specified by
            the `output_key` parameter of the last step.
        """
        x = {"main": input} if isinstance(input, torch.Tensor) else input
        for module in self:
            x = module(x, single_eval_pos=single_eval_pos, **kwargs)

        # For backwards compatibility
        if self.output_key is not None:
            return x[self.output_key]

        return x


class TorchPreprocessingStep(abc.ABC, nn.Module):
    """Abstract base class for sequential encoder steps.

    TorchPreprocessingStep is a wrapper around a module that defines the expected
    input keys and the produced output keys. The outputs are assigned to the output keys
    in the order specified by `out_keys`.

    Subclasses should create any state that depends on the train set in
    `_fit` and using it in `_transform`. This allows fitting on data first
    and doing inference later without refitting.
    """

    def __init__(
        self,
        in_keys: tuple[str, ...] = ("main",),
        out_keys: tuple[str, ...] = ("main",),
    ):
        """Init.

        Args:
            in_keys: The keys of the input tensors.
            out_keys: The keys to assign the output tensors to.
        """
        super().__init__()
        self.in_keys = in_keys
        self.out_keys = out_keys

    @abc.abstractmethod
    def _fit(
        self,
        state: dict[str, torch.Tensor],
        **kwargs: Any,
    ) -> None:
        """Fit the encoder step on the training set.

        Args:
            state: The dictionary containing the input tensors.
            single_eval_pos: The position to use for single evaluation.
            **kwargs: Additional keyword arguments passed to the encoder step.
        """
        ...

    @abc.abstractmethod
    def _transform(
        self,
        state: dict[str, torch.Tensor],
        **kwargs: Any,
    ) -> dict[str, torch.Tensor]:
        """Transform the data using the fitted encoder step.

        Args:
            state: The dictionary containing the input tensors.
            **kwargs: Additional keyword arguments passed to the encoder step.

        Returns:
            The transformed output dictionary.
        """
        ...

    def _validate_input_keys(self, state: dict) -> None:
        """Raise KeyError if expected input keys are missing from state."""
        missing = [k for k in self.in_keys if k not in state]
        if missing:
            raise KeyError(
                f"{self.__class__.__name__}: missing input tensor in dict `{missing}`. "
                f"Available keys in state: `{list(state.keys())}`"
            )

    def _validate_output_keys(self, outputs: dict) -> None:
        """Raise KeyError if unexpected output keys are present in outputs."""
        unexpected = set(outputs.keys()) - set(self.out_keys)
        if unexpected:
            raise KeyError(
                f"{self.__class__.__name__}: unexpected output tensor in dict "
                f"`{unexpected}`. Available keys in state: "
                f"`{list(outputs.keys())}`"
            )

        missing = [k for k in self.out_keys if k not in outputs]
        if missing:
            raise KeyError(
                f"{self.__class__.__name__}: missing output tensor in dict "
                f"`{missing}`. Available keys in state: `{list(outputs.keys())}`"
            )

    @override
    def forward(
        self,
        state: dict,
        *,
        single_eval_pos: int | None = None,
        cache_trainset_representation: bool = False,
        **kwargs: Any,
    ) -> dict:
        """Perform the forward pass of the encoder step.

        Args:
            state: The input state dictionary containing the input tensors.
            cache_trainset_representation:
                Whether to cache the training set representation. Only supported for
                _fit and _transform (not _forward).
            **kwargs: Additional keyword arguments passed to the encoder step.

        Returns:
            The updated state dictionary with the output tensor or a tuple of
            output tensors.
        """
        self._validate_input_keys(state)

        do_fit = (
            single_eval_pos is not None and single_eval_pos > 0
        ) or not cache_trainset_representation
        if do_fit:
            self._fit(state, single_eval_pos=single_eval_pos, **kwargs)
        outputs = self._transform(state, single_eval_pos=single_eval_pos, **kwargs)

        self._validate_output_keys(outputs)

        state.update(outputs)
        return state



"""Utility functions for preprocessing steps."""

from __future__ import annotations

from typing import TYPE_CHECKING
from typing_extensions import override

import numpy as np
from sklearn.impute import SimpleImputer
from sklearn.pipeline import Pipeline
from sklearn.preprocessing import FunctionTransformer, StandardScaler

if TYPE_CHECKING:
    from sklearn.base import TransformerMixin


def make_scaler_safe(name: str, scaler: TransformerMixin) -> Pipeline:
    """Wrap a scaler with steps that ensure all input and output values are finite.

    Inserts inf-to-nan conversion and mean imputation both before and after the
    scaler to guard against edge cases such as division-by-zero during scaling
    or non-finite values in the input data.

    Args:
        name: Name for the scaler step in the resulting pipeline.
        scaler: The scaler / transformer to wrap.

    Returns:
        A `Pipeline` with sanitization steps surrounding the scaler.
    """
    return Pipeline(
        steps=[*_make_finite_steps("pre"), (name, scaler), *_make_finite_steps("post")],
    )


def wrap_with_safe_standard_scaler(
    transformer: TransformerMixin,
) -> Pipeline:
    """Wrap a transformer with a safely-guarded `StandardScaler`.

    Useful for transformers like `PowerTransformer` that can produce inf
    values in edge cases, which would crash a subsequent `StandardScaler`.

    Args:
        transformer: The transformer to apply before standard scaling.

    Returns:
        A Pipeline of `[transformer_name, safe_standard_scaler]`.
    """
    return Pipeline(
        steps=[
            ("input_transformer", transformer),
            ("standard", make_scaler_safe("standard", StandardScaler())),
        ],
    )


class _NoInverseImputer(SimpleImputer):
    """SimpleImputer that returns input unchanged on `inverse_transform`."""

    @override
    def inverse_transform(self, X: np.ndarray) -> np.ndarray:
        return X


def _identity(x: np.ndarray) -> np.ndarray:
    """Return input unchanged. Used as a no-op inverse for FunctionTransformer."""
    return x


def _replace_inf_with_nan(x: np.ndarray) -> np.ndarray:
    """Replace +inf and -inf with NaN, leaving existing NaN values unchanged."""
    return np.nan_to_num(x, nan=np.nan, neginf=np.nan, posinf=np.nan)


def _make_finite_steps(suffix: str) -> list[tuple[str, TransformerMixin]]:
    """Create pipeline steps that replace non-finite values and impute NaNs.

    Args:
        suffix: Appended to step names to ensure uniqueness (e.g. "pre", "post").

    Returns:
        A list of `(name, transformer)` tuples for use as sklearn Pipeline steps.
    """
    return [
        (
            f"inf_to_nan_{suffix}",
            FunctionTransformer(
                func=_replace_inf_with_nan,
                inverse_func=_identity,
                check_inverse=False,
            ),
        ),
        (
            f"nan_impute_{suffix}",
            _NoInverseImputer(
                missing_values=np.nan,
                strategy="mean",
                # keep empty features so inverse_transform dimensions are consistent
                keep_empty_features=True,
            ),
        ),
    ]



"""Tests for TabPFN classifier finetuning functionality.

This module contains tests for:
- The FinetunedTabPFNClassifier wrapper class
- Dataset preprocessing utilities for finetuning
- Data collation and batching for training loops
- Checkpoint saving/loading functionality
"""

from __future__ import annotations

from collections.abc import Callable
from functools import partial
from pathlib import Path
from typing import Any, Literal
from unittest import mock
from unittest.mock import patch

import numpy as np
import pytest
import sklearn
import torch
from sklearn.datasets import make_classification
from sklearn.model_selection import train_test_split
from torch.utils.data import DataLoader

from tabpfn import TabPFNClassifier
from tabpfn.finetuning.data_util import (
    ClassifierBatch,
    DatasetCollectionWithPreprocessing,
    get_preprocessed_dataset_chunks,
    meta_dataset_collator,
)
from tabpfn.finetuning.finetuned_classifier import FinetunedTabPFNClassifier
from tabpfn.finetuning.train_util import get_checkpoint_path_and_epoch_from_output_dir
from tabpfn.preprocessing import ClassifierEnsembleConfig

from .utils import (
    get_pytest_devices,
    get_pytest_devices_with_mps_marked_slow,
    mark_mps_configs_as_slow,
)

rng = np.random.default_rng(42)

devices = get_pytest_devices()

FitMode = Literal["low_memory", "fit_preprocessors", "fit_with_cache", "batched"]

fit_modes: list[FitMode] = [
    "batched",
    "fit_preprocessors",
]

inference_precision_methods: list[torch.types._dtype | Literal["autocast", "auto"]] = [
    "auto",
    torch.float64,
]
estimators = [1, 2]

param_order = [
    "device",
    "n_estimators",
    "fit_mode",
    "inference_precision",
]

default_config: dict[str, Any] = {
    "n_estimators": 1,
    "device": "auto",
    "fit_mode": "batched",
    "inference_precision": "auto",
}

param_values: dict[str, list] = {
    "n_estimators": estimators,
    "device": devices,
    "fit_mode": fit_modes,
    "inference_precision": inference_precision_methods,
}

combinations = [tuple(default_config[param] for param in param_order)]
for param in param_order:
    for value in param_values[param]:
        if value != default_config[param]:
            config = default_config.copy()
            config[param] = value
            combinations.append(tuple(config[param] for param in param_order))


# =============================================================================
# Parametrization for FinetunedTabPFNClassifier tests
# =============================================================================

finetuned_param_order = [
    "device",
    "early_stopping",
    "use_lr_scheduler",
]

finetuned_default_config: dict[str, Any] = {
    "device": "cpu",
    "early_stopping": False,
    "use_lr_scheduler": False,
}

finetuned_param_values: dict[str, list] = {
    "device": devices,
    "early_stopping": [True, False],
    "use_lr_scheduler": [True, False],
}

# Generate combinations varying one parameter at a time from defaults
finetuned_combinations = [
    tuple(finetuned_default_config[param] for param in finetuned_param_order)
]
for param in finetuned_param_order:
    for value in finetuned_param_values[param]:
        if value != finetuned_default_config[param]:
            config = finetuned_default_config.copy()
            config[param] = value
            finetuned_combinations.append(
                tuple(config[param] for param in finetuned_param_order)
            )


def create_mock_architecture_forward(
    n_classes: int,
) -> Callable[..., torch.Tensor]:
    """Create a side_effect function for mocking the Architecture forward method.

    The Architecture.forward method signature is:
    forward(x, y, *, only_return_standard_out=True, categorical_inds=None)

    Where:
    - x has shape (train+test rows, batch_size, num_features)
    - y has shape (train rows, batch_size) or (train rows, batch_size, 1)
    - returns shape (test rows, batch_size, num_classes)

    Args:
        n_classes: Number of classes for the classification task.

    Returns:
        A mock forward function that returns random logits with correct shape.
    """

    def mock_forward(
        self: torch.nn.Module,
        x: torch.Tensor | dict[str, torch.Tensor],
        y: torch.Tensor | dict[str, torch.Tensor] | None,
        **_kwargs: bool,
    ) -> torch.Tensor:
        """Mock forward pass that returns random logits."""
        if isinstance(x, dict):
            x = x["main"]

        if y is not None:
            y_tensor = y["main"] if isinstance(y, dict) else y
            num_train_rows = y_tensor.shape[0]
        else:
            num_train_rows = 0

        # x has shape (train+test rows, batch_size, num_features)
        total_rows = x.shape[0]
        batch_size = x.shape[1]
        num_test_rows = total_rows - num_train_rows

        # Touch a model parameter so gradients flow during backward pass.
        # This is needed for GradScaler to record inf checks on CUDA
        # on older torch versions.
        first_param = next(self.parameters())
        param_contribution = 0.0 * first_param.sum()

        # Return shape (test_rows, batch_size, num_classes)
        return (
            torch.randn(
                num_test_rows,
                batch_size,
                n_classes,
                requires_grad=True,
                device=x.device,
            )
            + param_contribution
        )

    return mock_forward


@pytest.fixture(scope="module")
def synthetic_data() -> tuple[np.ndarray, np.ndarray]:
    """Generate synthetic data for testing."""
    X = rng.normal(size=(100, 4)).astype(np.float32)
    y = rng.integers(0, 3, size=100).astype(np.float32)
    return X, y


@pytest.fixture(scope="module")
def uniform_synthetic_dataset_collection() -> list[tuple[np.ndarray, np.ndarray]]:
    """Fixture: synthetic collection of datasets (list of (X, y) tuples)."""
    datasets = []
    for _ in range(3):
        X = rng.normal(size=(30, 3)).astype(np.float32)
        y = rng.integers(0, 3, size=30)
        datasets.append((X, y))
    return datasets


@pytest.fixture(scope="module")
def classification_data() -> tuple[np.ndarray, np.ndarray, np.ndarray, np.ndarray]:
    """Generate simple classification data."""
    X, y = make_classification(
        n_samples=20,
        n_features=5,
        n_informative=3,
        n_redundant=0,
        n_classes=3,
        random_state=42,
    )
    y = np.asarray(y)
    y = y - y.min()
    X_train, X_test, y_train, y_test = train_test_split(
        X, y, test_size=0.3, random_state=42, stratify=y
    )
    return (
        np.asarray(X_train),
        np.asarray(X_test),
        np.asarray(y_train),
        np.asarray(y_test),
    )


@pytest.fixture(params=devices)
def classifier_instance(request: pytest.FixtureRequest) -> TabPFNClassifier:
    """Provides a basic classifier instance, parameterized by device."""
    device = request.param
    if device == "cuda" and not torch.cuda.is_available():
        pytest.skip("CUDA device requested but not available.")
    return TabPFNClassifier(
        n_estimators=2,
        device=device,
        random_state=42,
        inference_precision=torch.float32,
        fit_mode="batched",
        differentiable_input=False,
    )


def create_classifier(
    n_estimators: int,
    device: str,
    fit_mode: FitMode,
    inference_precision: torch.types._dtype | Literal["autocast", "auto"],
    **kwargs: Any,
) -> TabPFNClassifier:
    """Instantiates classifier with common parameters."""
    if device == "cpu" and inference_precision == "autocast":
        pytest.skip("Unsupported combination: CPU with 'autocast'")
    if device == "cuda" and not torch.cuda.is_available():
        pytest.skip("CUDA device requested but not available.")

    default_kwargs: dict[str, Any] = {"random_state": 42}
    default_kwargs.update(kwargs)

    return TabPFNClassifier(
        n_estimators=n_estimators,
        device=device,
        fit_mode=fit_mode,
        inference_precision=inference_precision,
        **default_kwargs,
    )


@pytest.fixture
def variable_synthetic_dataset_collection() -> list[tuple[np.ndarray, np.ndarray]]:
    """Fixture: synthetic collection of datasets with varying sizes and classes."""
    datasets = []
    dataset_sizes = [10, 20, 30]
    class_counts = [2, 4, 6]
    n_features = 3
    for size, n_classes in zip(dataset_sizes, class_counts):
        X = rng.normal(size=(size, n_features)).astype(np.float32)
        y = rng.integers(0, n_classes, size=size)
        datasets.append((X, y))
    return datasets


# =============================================================================
# Tests for FinetunedTabPFNClassifier (Training Loop)
# =============================================================================


@pytest.mark.parametrize(
    ("device", "early_stopping", "use_lr_scheduler"),
    mark_mps_configs_as_slow(finetuned_combinations),
)
def test_finetuned_tabpfn_classifier_fit_and_predict(
    device: str,
    early_stopping: bool,
    use_lr_scheduler: bool,
    synthetic_data: tuple[np.ndarray, np.ndarray],
) -> None:
    """Test FinetunedTabPFNClassifier with various parameter combinations.

    Uses mocked Architecture.forward for fast test execution.
    Tests combinations of device, early_stopping, use_lr_scheduler,
    and n_estimators_finetune.
    """
    if device == "cuda" and not torch.cuda.is_available():
        pytest.skip("CUDA device requested but not available.")

    X, y = synthetic_data
    n_classes = len(np.unique(y))
    X_train, X_test, y_train, _ = train_test_split(X, y, test_size=0.3, random_state=42)
    X_train = np.asarray(X_train)
    X_test = np.asarray(X_test)
    y_train = np.asarray(y_train)

    epochs = 4 if early_stopping else 2

    finetuned_clf = FinetunedTabPFNClassifier(
        device=device,
        epochs=epochs,
        learning_rate=1e-4,
        validation_split_ratio=0.2,
        n_finetune_ctx_plus_query_samples=50,
        finetune_ctx_query_split_ratio=0.1,
        n_inference_subsample_samples=100,
        random_state=42,
        early_stopping=early_stopping,
        early_stopping_patience=2,
        n_estimators_finetune=1,
        n_estimators_validation=1,
        n_estimators_final_inference=1,
        use_lr_scheduler=use_lr_scheduler,
        lr_warmup_only=False,
    )

    mock_forward = create_mock_architecture_forward(n_classes=n_classes)

    with mock.patch(
        "tabpfn.architectures.base.transformer.PerFeatureTransformer.forward",
        autospec=True,
        side_effect=mock_forward,
    ):
        finetuned_clf.fit(X_train, y_train)

    assert finetuned_clf.is_fitted_
    assert hasattr(finetuned_clf, "finetuned_estimator_")
    assert hasattr(finetuned_clf, "finetuned_inference_classifier_")

    probabilities = finetuned_clf.predict_proba(X_test)
    assert probabilities.shape[0] == X_test.shape[0]
    assert probabilities.shape[1] == n_classes
    assert np.allclose(probabilities.sum(axis=1), 1.0, atol=1e-5)

    predictions = finetuned_clf.predict(X_test)
    assert predictions.shape[0] == X_test.shape[0]
    assert all(pred in np.unique(y_train) for pred in predictions)


# =============================================================================
# Tests for Checkpoint Saving and Loading
# =============================================================================


@pytest.mark.parametrize("device", get_pytest_devices_with_mps_marked_slow())
def test_checkpoint_saving_and_loading(
    device: str,
    tmp_path: Path,
    synthetic_data: tuple[np.ndarray, np.ndarray],
) -> None:
    """Test that checkpoints are saved and can be loaded correctly."""
    if device == "cuda" and not torch.cuda.is_available():
        pytest.skip("CUDA device requested but not available.")

    X, y = synthetic_data
    n_classes = len(np.unique(y))
    X_train, _, y_train, _ = train_test_split(X, y, test_size=0.3, random_state=42)
    X_train = np.asarray(X_train)
    y_train = np.asarray(y_train)

    output_folder = tmp_path / "checkpoints"

    finetuned_clf = FinetunedTabPFNClassifier(
        device=device,
        epochs=4,
        learning_rate=1e-5,
        validation_split_ratio=0.2,
        n_finetune_ctx_plus_query_samples=50,
        finetune_ctx_query_split_ratio=0.1,
        n_inference_subsample_samples=100,
        random_state=42,
        early_stopping=False,
        use_lr_scheduler=False,
        n_estimators_finetune=1,
        n_estimators_validation=1,
        n_estimators_final_inference=1,
        save_checkpoint_interval=2,
    )

    mock_forward = create_mock_architecture_forward(n_classes=n_classes)

    with mock.patch(
        "tabpfn.architectures.base.transformer.PerFeatureTransformer.forward",
        autospec=True,
        side_effect=mock_forward,
    ):
        finetuned_clf.fit(X_train, y_train, output_dir=output_folder)

    # Verify interval checkpoints exist
    checkpoint_files = sorted(output_folder.glob("checkpoint_*_[0-9]*.pth"))
    assert len(checkpoint_files) >= 2, (
        "At least two interval checkpoints should be saved"
    )

    # Check specific epoch checkpoints
    epoch_2_checkpoint = output_folder / "checkpoint_70_2.pth"
    assert epoch_2_checkpoint.exists(), "Checkpoint at epoch 2 should exist"

    epoch_4_checkpoint = output_folder / "checkpoint_70_4.pth"
    assert epoch_4_checkpoint.exists(), "Checkpoint at epoch 4 should exist"

    # Verify best checkpoint exists
    best_checkpoint_path = output_folder / "checkpoint_70_best.pth"
    assert best_checkpoint_path.exists(), "Best checkpoint should exist"

    # Verify checkpoint structure
    checkpoint = torch.load(epoch_2_checkpoint, weights_only=False)
    assert "state_dict" in checkpoint, "Checkpoint should contain state_dict"
    assert "config" in checkpoint, "Checkpoint should contain config"
    assert "optimizer" in checkpoint, "Checkpoint should contain optimizer state"
    assert "epoch" in checkpoint, "Checkpoint should contain epoch number"
    assert "roc_auc" in checkpoint, "Checkpoint should contain roc_auc"
    assert "log_loss" in checkpoint, "Checkpoint should contain log_loss"
    assert checkpoint["epoch"] == 2, "Checkpoint should be from epoch 2"

    # Verify best checkpoint structure
    best_checkpoint = torch.load(best_checkpoint_path, weights_only=False)
    assert "state_dict" in best_checkpoint
    assert "epoch" in best_checkpoint
    assert isinstance(best_checkpoint["epoch"], int)
    assert best_checkpoint["epoch"] > 0


@pytest.mark.parametrize("device", get_pytest_devices_with_mps_marked_slow())
def test_checkpoint_resumption(
    device: str,
    tmp_path: Path,
    synthetic_data: tuple[np.ndarray, np.ndarray],
) -> None:
    """Test that training can be resumed from a checkpoint."""
    if device == "cuda" and not torch.cuda.is_available():
        pytest.skip("CUDA device requested but not available.")

    X, y = synthetic_data
    n_classes = len(np.unique(y))
    X_train, X_test, y_train, _ = train_test_split(X, y, test_size=0.3, random_state=42)
    X_train = np.asarray(X_train)
    X_test = np.asarray(X_test)
    y_train = np.asarray(y_train)

    output_folder = tmp_path / "checkpoints_resume"

    # Train for 2 epochs with checkpoint_interval=2
    finetuned_clf = FinetunedTabPFNClassifier(
        device=device,
        epochs=2,
        learning_rate=1e-5,
        validation_split_ratio=0.2,
        n_finetune_ctx_plus_query_samples=50,
        finetune_ctx_query_split_ratio=0.1,
        n_inference_subsample_samples=100,
        random_state=42,
        early_stopping=False,
        use_lr_scheduler=False,
        n_estimators_finetune=1,
        n_estimators_validation=1,
        n_estimators_final_inference=1,
        save_checkpoint_interval=2,
    )

    mock_forward = create_mock_architecture_forward(n_classes=n_classes)

    with mock.patch(
        "tabpfn.architectures.base.transformer.PerFeatureTransformer.forward",
        autospec=True,
        side_effect=mock_forward,
    ):
        finetuned_clf.fit(X_train, y_train, output_dir=output_folder)

    epoch_2_checkpoint = output_folder / "checkpoint_70_2.pth"
    assert epoch_2_checkpoint.exists(), "Checkpoint at epoch 2 should exist"

    best_checkpoint_path = output_folder / "checkpoint_70_best.pth"
    assert best_checkpoint_path.exists(), (
        "Best checkpoint should exist after first training"
    )

    # Resume training for another 2 epochs (total 4)
    finetuned_clf_resumed = FinetunedTabPFNClassifier(
        device=device,
        epochs=4,  # Total epochs = 4
        learning_rate=1e-5,
        validation_split_ratio=0.2,
        n_finetune_ctx_plus_query_samples=50,
        finetune_ctx_query_split_ratio=0.1,
        n_inference_subsample_samples=100,
        random_state=42,
        early_stopping=False,
        use_lr_scheduler=False,
        n_estimators_finetune=1,
        n_estimators_validation=1,
        n_estimators_final_inference=1,
        save_checkpoint_interval=2,
    )

    with mock.patch(
        "tabpfn.architectures.base.transformer.PerFeatureTransformer.forward",
        autospec=True,
        side_effect=mock_forward,
    ):
        finetuned_clf_resumed.fit(X_train, y_train, output_dir=output_folder)

    epoch_4_checkpoint = output_folder / "checkpoint_70_4.pth"
    assert epoch_4_checkpoint.exists(), (
        "Checkpoint at epoch 4 should exist after resumption"
    )

    assert best_checkpoint_path.exists()

    # Verify predictions work
    y_pred_proba = finetuned_clf_resumed.predict_proba(X_test)
    y_pred = finetuned_clf_resumed.predict(X_test)

    assert y_pred_proba.shape == (len(X_test), n_classes)
    assert y_pred.shape == (len(X_test),)
    assert finetuned_clf_resumed.is_fitted_


def test_checkpoint_epoch_offset_extraction(tmp_path: Path) -> None:
    """Test that epoch offset is correctly extracted from checkpoint filenames."""
    # Test with no checkpoints
    output_folder = tmp_path / "no_checkpoints"
    output_folder.mkdir(parents=True, exist_ok=True)
    model_path, epoch = get_checkpoint_path_and_epoch_from_output_dir(
        output_folder, train_size=100
    )
    assert model_path is None
    assert epoch == 0

    # Test with one checkpoint file
    output_folder_with_checkpoint = tmp_path / "with_checkpoint"
    output_folder_with_checkpoint.mkdir(parents=True, exist_ok=True)
    checkpoint_path = output_folder_with_checkpoint / "checkpoint_100_10.pth"
    torch.save({"dummy": "data"}, checkpoint_path)

    model_path, epoch = get_checkpoint_path_and_epoch_from_output_dir(
        output_folder_with_checkpoint, train_size=100, get_best=False
    )
    assert model_path == checkpoint_path
    assert epoch == 10

    # Test with multiple checkpoint files (should return the last one)
    checkpoint_path_20 = output_folder_with_checkpoint / "checkpoint_100_20.pth"
    torch.save({"dummy": "data"}, checkpoint_path_20)

    model_path, epoch = get_checkpoint_path_and_epoch_from_output_dir(
        output_folder_with_checkpoint, train_size=100, get_best=False
    )
    assert model_path == checkpoint_path_20
    assert epoch == 20

    # Test with best checkpoint (should prioritize best over numbered)
    best_checkpoint_path = output_folder_with_checkpoint / "checkpoint_100_best.pth"
    torch.save({"epoch": 15, "dummy": "data"}, best_checkpoint_path)

    model_path, epoch = get_checkpoint_path_and_epoch_from_output_dir(
        output_folder_with_checkpoint, train_size=100, get_best=True
    )
    assert model_path == best_checkpoint_path
    assert epoch == 15


@pytest.mark.parametrize("device", get_pytest_devices_with_mps_marked_slow())
def test_checkpoint_interval_configuration(
    device: str,
    tmp_path: Path,
    synthetic_data: tuple[np.ndarray, np.ndarray],
) -> None:
    """Test that checkpoint_interval parameter works correctly."""
    if device == "cuda" and not torch.cuda.is_available():
        pytest.skip("CUDA device requested but not available.")

    X, y = synthetic_data
    n_classes = len(np.unique(y))
    X_train, _, y_train, _ = train_test_split(X, y, test_size=0.3, random_state=42)
    X_train = np.asarray(X_train)
    y_train = np.asarray(y_train)

    output_folder = tmp_path / "checkpoint_interval_test"

    # Train for 6 epochs with checkpoint_interval=3
    finetuned_clf = FinetunedTabPFNClassifier(
        device=device,
        epochs=6,
        learning_rate=1e-5,
        validation_split_ratio=0.2,
        n_finetune_ctx_plus_query_samples=50,
        finetune_ctx_query_split_ratio=0.1,
        n_inference_subsample_samples=100,
        random_state=42,
        early_stopping=False,
        use_lr_scheduler=False,
        n_estimators_finetune=1,
        n_estimators_validation=1,
        n_estimators_final_inference=1,
        save_checkpoint_interval=3,
    )

    mock_forward = create_mock_architecture_forward(n_classes=n_classes)

    with mock.patch(
        "tabpfn.architectures.base.transformer.PerFeatureTransformer.forward",
        autospec=True,
        side_effect=mock_forward,
    ):
        finetuned_clf.fit(X_train, y_train, output_dir=output_folder)

    # Verify checkpoints saved at epochs 3, 6
    epoch_3_checkpoint = output_folder / "checkpoint_70_3.pth"
    epoch_6_checkpoint = output_folder / "checkpoint_70_6.pth"

    assert epoch_3_checkpoint.exists(), "Checkpoint at epoch 3 should exist"
    assert epoch_6_checkpoint.exists(), "Checkpoint at epoch 6 should exist"

    # Verify no checkpoints at epochs 1, 2, 4, 5
    for epoch in [1, 2, 4, 5]:
        checkpoint_path = output_folder / f"checkpoint_70_{epoch}.pth"
        assert not checkpoint_path.exists(), (
            f"Checkpoint at epoch {epoch} should not exist"
        )

    # Verify best checkpoint exists
    best_checkpoint_path = output_folder / "checkpoint_70_best.pth"
    assert best_checkpoint_path.exists(), "Best checkpoint should exist"


@pytest.mark.parametrize("device", get_pytest_devices_with_mps_marked_slow())
def test_best_checkpoint_saving(
    device: str,
    tmp_path: Path,
    synthetic_data: tuple[np.ndarray, np.ndarray],
) -> None:
    """Test that best checkpoint is saved correctly."""
    if device == "cuda" and not torch.cuda.is_available():
        pytest.skip("CUDA device requested but not available.")

    X, y = synthetic_data
    n_classes = len(np.unique(y))
    X_train, _, y_train, _ = train_test_split(X, y, test_size=0.3, random_state=42)
    X_train = np.asarray(X_train)
    y_train = np.asarray(y_train)

    output_folder = tmp_path / "best_checkpoint_test"

    # Train for 3 epochs with checkpoint_interval=None (no interval saves)
    finetuned_clf = FinetunedTabPFNClassifier(
        device=device,
        epochs=3,
        learning_rate=1e-5,
        validation_split_ratio=0.2,
        n_finetune_ctx_plus_query_samples=50,
        finetune_ctx_query_split_ratio=0.1,
        n_inference_subsample_samples=100,
        random_state=42,
        early_stopping=False,
        use_lr_scheduler=False,
        n_estimators_finetune=1,
        n_estimators_validation=1,
        n_estimators_final_inference=1,
        save_checkpoint_interval=None,  # Disable interval checkpoints
    )

    mock_forward = create_mock_architecture_forward(n_classes=n_classes)

    with mock.patch(
        "tabpfn.architectures.base.transformer.PerFeatureTransformer.forward",
        autospec=True,
        side_effect=mock_forward,
    ):
        finetuned_clf.fit(X_train, y_train, output_dir=output_folder)

    # Verify that best checkpoint exists
    best_checkpoint_path = output_folder / "checkpoint_70_best.pth"
    assert best_checkpoint_path.exists(), "Best checkpoint should exist"

    # Verify no interval checkpoints exist (since checkpoint_interval=None)
    interval_checkpoints = list(output_folder.glob("checkpoint_70_[0-9]*.pth"))
    assert len(interval_checkpoints) == 0, (
        "No interval checkpoints should exist when checkpoint_interval=None"
    )


# =============================================================================
# Tests for Dataset Preprocessing Utilities
# =============================================================================


def test_get_preprocessed_datasets_basic() -> None:
    """Test basic functionality of get_preprocessed_datasets_helper."""
    X = rng.normal(size=(100, 4)).astype(np.float32)
    y = rng.integers(0, 3, size=100)

    clf = TabPFNClassifier()
    dataset = get_preprocessed_dataset_chunks(
        clf,
        X,
        y,
        train_test_split,
        100,
        model_type="classifier",
        equal_split_size=True,
        seed=42,
    )
    assert hasattr(dataset, "__getitem__")
    assert hasattr(dataset, "__len__")
    assert len(dataset) > 0
    item = dataset[0]
    assert isinstance(item, ClassifierBatch)
    assert hasattr(item, "X_context")
    assert hasattr(item, "X_query")
    assert hasattr(item, "y_context")
    assert hasattr(item, "y_query")
    assert hasattr(item, "cat_indices")
    assert hasattr(item, "configs")


def test_datasetcollectionwithpreprocessing_classification_single_dataset(
    synthetic_data: tuple[np.ndarray, np.ndarray],
    classifier_instance: TabPFNClassifier,
) -> None:
    """Test DatasetCollectionWithPreprocessing with a single dataset."""
    X_raw, y_raw = synthetic_data
    clf = classifier_instance
    n_estimators = clf.n_estimators
    test_size = 0.3

    split_fn = partial(train_test_split, test_size=test_size, shuffle=True)
    dataset_collection = get_preprocessed_dataset_chunks(
        classifier_instance,
        X_raw,
        y_raw,
        split_fn=split_fn,
        max_data_size=None,
        model_type="classifier",
        equal_split_size=True,
        seed=42,
    )

    assert isinstance(dataset_collection, DatasetCollectionWithPreprocessing)
    assert len(dataset_collection) == 1, "Collection should contain one dataset config"

    item_index = 0
    batch = dataset_collection[item_index]

    assert isinstance(batch, ClassifierBatch)

    assert isinstance(batch.X_context, list)
    assert len(batch.X_context) == n_estimators
    n_samples_total = X_raw.shape[0]
    expected_n_test = int(np.floor(n_samples_total * test_size))
    expected_n_train = n_samples_total - expected_n_test
    assert batch.y_query.shape == (expected_n_test,)
    assert batch.X_context[0].shape[0] == expected_n_train


def test_datasetcollectionwithpreprocessing_classification_multiple_datasets(
    uniform_synthetic_dataset_collection: list[tuple[np.ndarray, np.ndarray]],
    classifier_instance: TabPFNClassifier,
) -> None:
    """Test DatasetCollectionWithPreprocessing using multiple synthetic datasets."""
    datasets = uniform_synthetic_dataset_collection
    clf = classifier_instance
    n_estimators = clf.n_estimators
    test_size = 0.3
    split_fn = partial(train_test_split, test_size=test_size, shuffle=True)

    X_list = [X for X, _ in datasets]
    y_list = [y for _, y in datasets]

    dataset_collection = get_preprocessed_dataset_chunks(
        clf,
        X_list,
        y_list,
        split_fn=split_fn,
        max_data_size=None,
        model_type="classifier",
        equal_split_size=True,
        seed=42,
    )

    assert isinstance(dataset_collection, DatasetCollectionWithPreprocessing)
    assert len(dataset_collection) == len(datasets), (
        "Collection should contain one item per dataset"
    )

    for item_index in range(len(datasets)):
        batch = dataset_collection[item_index]
        assert isinstance(batch, ClassifierBatch)
        assert isinstance(batch.X_context, list)
        assert len(batch.X_context) == n_estimators
        n_samples_total = X_list[item_index].shape[0]
        expected_n_test = int(np.floor(n_samples_total * test_size))
        expected_n_train = n_samples_total - expected_n_test
        assert batch.y_query.shape == (expected_n_test,)
        assert batch.X_context[0].shape[0] == expected_n_train


def test_dataset_and_collator_with_dataloader_uniform(
    uniform_synthetic_dataset_collection: list[tuple[np.ndarray, np.ndarray]],
    classifier_instance: TabPFNClassifier,
) -> None:
    """Test dataset and collator integration with DataLoader using uniform data."""
    X_list = [X for X, _ in uniform_synthetic_dataset_collection]
    y_list = [y for _, y in uniform_synthetic_dataset_collection]
    dataset_collection = get_preprocessed_dataset_chunks(
        classifier_instance,
        X_list,
        y_list,
        train_test_split,
        100,
        model_type="classifier",
        equal_split_size=True,
        seed=42,
    )
    batch_size = 1
    dl = DataLoader(
        dataset_collection,
        batch_size=batch_size,
        collate_fn=meta_dataset_collator,
    )
    for batch in dl:
        assert isinstance(batch, ClassifierBatch)
        for est_tensor in batch.X_context:
            assert isinstance(est_tensor, torch.Tensor), (
                "Each estimator's batch should be a tensor."
            )
            assert est_tensor.shape[0] == batch_size
        for est_tensor in batch.y_context:
            assert isinstance(est_tensor, torch.Tensor), (
                "Each estimator's batch should be a tensor for labels."
            )
            assert est_tensor.shape[0] == batch_size
        break  # Only check one batch


def test_classifier_dataset_and_collator_batches_type(
    variable_synthetic_dataset_collection: list[tuple[np.ndarray, np.ndarray]],
    classifier_instance: TabPFNClassifier,
) -> None:
    """Test that batches from dataset and collator have correct types."""
    X_list = [X for X, _ in variable_synthetic_dataset_collection]
    y_list = [y for _, y in variable_synthetic_dataset_collection]
    dataset_collection = get_preprocessed_dataset_chunks(
        classifier_instance,
        X_list,
        y_list,
        train_test_split,
        100,
        model_type="classifier",
        equal_split_size=True,
        seed=42,
    )
    batch_size = 1
    dl = DataLoader(
        dataset_collection,
        batch_size=batch_size,
        collate_fn=meta_dataset_collator,
    )
    for batch in dl:
        assert isinstance(batch, ClassifierBatch)
        for est_tensor in batch.X_context:
            assert isinstance(est_tensor, torch.Tensor)
            assert est_tensor.shape[0] == batch_size
        for est_tensor in batch.y_context:
            assert isinstance(est_tensor, torch.Tensor)
            assert est_tensor.shape[0] == batch_size
        assert isinstance(batch.cat_indices, list)
        for conf in batch.configs:
            for c in conf:
                assert isinstance(c, ClassifierEnsembleConfig)
        break


def test_get_preprocessed_datasets_multiple_datasets(
    classifier_instance: TabPFNClassifier,
) -> None:
    """Test get_preprocessed_datasets_helper with multiple datasets."""
    X1 = rng.standard_normal((10, 4))
    y1 = rng.integers(0, 2, size=10)
    X2 = rng.standard_normal((8, 4))
    y2 = rng.integers(0, 2, size=8)
    datasets = get_preprocessed_dataset_chunks(
        classifier_instance,
        [X1, X2],
        [y1, y2],
        split_fn=lambda x, y: (x, y),
        max_data_size=10_000,
        model_type="classifier",
        equal_split_size=True,
        seed=42,
    )
    assert hasattr(datasets, "__getitem__")
    assert len(datasets) == 2


def test_get_preprocessed_datasets_categorical_features(
    classifier_instance: TabPFNClassifier,
) -> None:
    """Test get_preprocessed_datasets_helper with categorical features."""
    X = np.array([[0, 1.2, 3.4], [1, 2.3, 4.5], [0, 0.1, 2.2], [2, 1.1, 3.3]])
    y = np.array([0, 1, 0, 1])
    classifier_instance.categorical_features_indices = [0]
    datasets = get_preprocessed_dataset_chunks(
        classifier_instance,
        X,
        y,
        split_fn=lambda x, y: (x, y),
        model_type="classifier",
        max_data_size=None,
        equal_split_size=True,
        seed=42,
    )
    assert hasattr(datasets, "__getitem__")


def test_forward_runs(
    classifier_instance: TabPFNClassifier,
    classification_data: tuple[np.ndarray, np.ndarray, np.ndarray, np.ndarray],
) -> None:
    """Ensure forward runs OK after standard fit."""
    X_train, X_test, y_train, _y_test = classification_data
    clf = classifier_instance
    clf.fit_mode = "low_memory"
    clf.fit(X_train, y_train)
    preds = clf.forward(
        torch.tensor(X_test, dtype=torch.float32), use_inference_mode=True
    )
    assert preds.ndim == 2, f"Expected 2D output, got {preds.shape}"
    assert preds.shape[0] == X_test.shape[0], "Mismatch in test sample count"
    assert preds.shape[1] == clf.n_classes_, "Mismatch in class count"
    probs_sum = preds.sum(dim=1)
    assert torch.allclose(probs_sum, torch.ones_like(probs_sum), atol=1e-5), (
        "Probabilities do not sum to 1"
    )


def test_fit_from_preprocessed_runs(
    classifier_instance: TabPFNClassifier,
    classification_data: tuple[np.ndarray, np.ndarray, np.ndarray, np.ndarray],
) -> None:
    """Verify fit_from_preprocessed runs and produces valid predictions."""
    X_train, _X_test, y_train, _y_test = classification_data
    clf = classifier_instance

    split_fn = partial(train_test_split, test_size=0.3, random_state=42)

    datasets_list = get_preprocessed_dataset_chunks(
        clf,
        X_train,
        y_train,
        split_fn,
        100,
        model_type="classifier",
        equal_split_size=True,
        seed=42,
    )
    batch_size = 1
    dl = DataLoader(
        datasets_list, batch_size=batch_size, collate_fn=meta_dataset_collator
    )

    for batch in dl:
        assert isinstance(batch, ClassifierBatch)
        clf.fit_from_preprocessed(
            batch.X_context, batch.y_context, batch.cat_indices, batch.configs
        )
        preds = clf.forward(batch.X_query)
        assert preds.ndim == 3, f"Expected 3D output, got {preds.shape}"
        assert preds.shape[0] == batch.X_query[0].shape[0]
        assert preds.shape[0] == batch.y_query.shape[0]
        assert preds.shape[1] == clf.n_classes_

        probs_sum = preds.sum(dim=1)
        assert torch.allclose(probs_sum, torch.ones_like(probs_sum), atol=1e-5), (
            "Probabilities do not sum to 1"
        )
        break


def test_finetuning_consistency_preprocessing_classifier() -> None:
    """Test consistency between standard and finetuning preprocessing pipelines.

    Compares tensors entering the internal model for:
    - Standard fit -> predict_proba path
    - Batched fit_from_preprocessed -> forward path
    """
    test_set_size = 0.3
    common_seed = 42
    n_total = 50  # Increased slightly for more robust testing
    n_features = 8
    n_classes = 2  # Use a specific number of classes
    n_informative = 5  # For make_classification
    n_estimators = 1  # Keep N=1 for easier direct comparison of tensors

    # --- 1. Setup ---
    X, y = sklearn.datasets.make_classification(
        n_samples=n_total,
        n_features=n_features,
        n_informative=n_informative,
        n_redundant=n_features - n_informative,
        n_classes=n_classes,
        n_clusters_per_class=1,
        random_state=common_seed,
    )
    splitfn = partial(
        train_test_split,
        test_size=test_set_size,
        random_state=common_seed,
        shuffle=False,  # Keep False for consistent splitting
    )
    X_train_raw, X_test_raw, y_train_raw, _y_test_raw = splitfn(X, y)

    # Initialize two classifiers with the necessary modes
    clf_standard = TabPFNClassifier(
        n_estimators=n_estimators,
        device="auto",
        random_state=common_seed,
        fit_mode="fit_preprocessors",  # A standard mode that preprocesses on fit
    )
    # 'batched' mode is required for get_preprocessed_datasets
    #  and fit_from_preprocessed
    clf_batched = TabPFNClassifier(
        n_estimators=n_estimators,
        device="auto",
        random_state=common_seed,
        fit_mode="batched",
    )

    # --- 2. Path 1: Standard fit -> predict_proba -> Capture Tensor ---

    clf_standard.fit(X_train_raw, y_train_raw)
    # Ensure the internal model attribute exists after fit
    assert hasattr(clf_standard, "models_"), (
        "Standard classifier models_ not found after fit."
    )
    assert hasattr(clf_standard.models_[0], "forward"), (
        "Standard classifier models_[0].forward not found after fit."
    )

    tensor_p1_full = None
    # Patch the standard classifier's *internal model's* forward method
    # The internal model typically receives the combined train+test sequence
    with patch.object(
        clf_standard.models_[0], "forward", wraps=clf_standard.models_[0].forward
    ) as mock_forward_p1:
        _ = clf_standard.predict_proba(X_test_raw)
        assert mock_forward_p1.called, "Standard models_[0].forward was not called."

        # Capture the tensor input 'x' (usually the second positional argument)
        call_args_list = mock_forward_p1.call_args_list
        assert len(call_args_list) > 0, (
            "No calls recorded for standard models_[0].forward."
        )
        assert len(call_args_list[0].args) > 1, (
            f"Standard models_[0].forward call had "
            f"unexpected arguments: {call_args_list[0].args}"
        )
        tensor_p1_full = mock_forward_p1.call_args.args[0]

    assert tensor_p1_full is not None, "Failed to capture tensor from standard path."
    # Shape might be [1, N_Total, Features+1] or similar. Check the actual shape.
    # Example assertion: Check if the sequence length matches n_total
    assert tensor_p1_full.shape[0] == n_total, (
        f"Path 1 tensor sequence length ({tensor_p1_full.shape[0]})"
        f"does not match n_total ({n_total}). Shape was {tensor_p1_full.shape}"
    )

    # FT Workflow (get_prep -> fit_prep -> predict_prep -> Capture Tensor) ---
    # Step 3a: Get preprocessed datasets using the *full* dataset
    # Requires fit_mode='batched' on clf_batched
    # Make sure default max_data_size is large enough.
    datasets_list = get_preprocessed_dataset_chunks(
        clf_batched,
        X,
        y,
        split_fn=splitfn,  # Use the full X, y and the split function
        max_data_size=10_000,
        model_type="classifier",
        equal_split_size=True,
        seed=42,
        force_no_stratify=True,
        shuffle=False,
    )
    assert len(datasets_list) > 0, "get_preprocessed_datasets returned empty list."

    dataloader = DataLoader(
        datasets_list,
        batch_size=1,
        collate_fn=meta_dataset_collator,
        shuffle=False,
    )
    batch = next(iter(dataloader), None)
    assert batch is not None, "DataLoader yielded no batches."
    assert isinstance(batch, ClassifierBatch)

    clf_batched.fit_from_preprocessed(
        batch.X_context, batch.y_context, batch.cat_indices, batch.configs
    )
    assert hasattr(clf_batched, "models_"), (
        "Batched classifier models_ not found after fit_from_preprocessed."
    )
    assert hasattr(clf_batched.models_[0], "forward"), (
        "Batched classifier models_[0].forward not found after fit_from_preprocessed."
    )

    # Step 3c: Call forward and capture the input tensor
    # to the *internal transformer model*
    tensor_p2_full = None
    # Patch the *batched* classifier's internal model's forward method
    with patch.object(
        clf_batched.models_[0], "forward", wraps=clf_batched.models_[0].forward
    ) as mock_forward_p2:
        _ = clf_batched.forward(batch.X_query)
        assert mock_forward_p2.called, "Batched models_[0].forward was not called."

        # Capture the tensor input 'x' (assuming same argument position as Path 1)
        call_args_list = mock_forward_p2.call_args_list
        assert len(call_args_list) > 0, (
            "No calls recorded for batched models_[0].forward."
        )
        assert len(call_args_list[0].args) > 1, (
            f"Batched models_[0].forward call had "
            f"unexpected arguments: {call_args_list[0].args}"
        )
        tensor_p2_full = mock_forward_p2.call_args.args[0]

    assert tensor_p2_full is not None, "Failed to capture tensor from batched path."
    # The internal model in this path should
    # also receive the full sequence if n_estimators=1
    # and the dataloader yielded the full split.
    assert tensor_p2_full.shape[0] == n_total, (
        f"Path 2 tensor sequence length ({tensor_p2_full.shape[0]}) "
        f"does not match n_total ({n_total}). Shape was {tensor_p2_full.shape}"
    )

    # --- 4. Comparison (Path 1 vs Path 2) ---

    # Ensure tensors are on the same device (CPU) for comparison
    tensor_p1_full = tensor_p1_full.cpu().squeeze(0)
    tensor_p2_full = tensor_p2_full.cpu().squeeze(0)

    # Final check of shapes after potential squeeze
    assert tensor_p1_full.shape == tensor_p2_full.shape, (
        "Shapes of final model input tensors mismatch after squeeze."
    )

    # Perform numerical comparison using torch.allclose
    # Use a reasonably small tolerance. Preprocessing should be near-identical.
    # Floating point ops might introduce tiny differences.
    atol = 1e-6
    rtol = 1e-5
    tensors_match = torch.allclose(tensor_p1_full, tensor_p2_full, atol=atol, rtol=rtol)

    if not tensors_match:
        diff = torch.abs(tensor_p1_full - tensor_p2_full)
        # Find where they differ most
        _max_diff_val, max_diff_idx = torch.max(diff.flatten(), dim=0)
        np.unravel_index(max_diff_idx.item(), tensor_p1_full.shape)

    assert tensors_match, "Mismatch between final model input tensors."



"""Abstract base class for fine-tuning TabPFN models.

This module provides the FinetunedTabPFNBase class, which contains shared
functionality for fine-tuning TabPFN on a specific dataset using the familiar
scikit-learn .fit() and .predict() API.
"""

from __future__ import annotations

import copy
import logging
import time
import warnings
from abc import ABC, abstractmethod
from dataclasses import dataclass, field
from functools import partial
from pathlib import Path
from typing import TYPE_CHECKING, Any, Literal

import numpy as np
import torch
from sklearn.base import BaseEstimator
from sklearn.model_selection import train_test_split
from torch.nn.utils import clip_grad_norm_
from torch.optim.lr_scheduler import LambdaLR
from torch.utils.data import DataLoader
from tqdm.auto import tqdm

from tabpfn.finetuning._torch_compat import GradScaler, autocast, sdpa_kernel_context
from tabpfn.finetuning.data_util import (
    ClassifierBatch,
    RegressorBatch,
    get_preprocessed_dataset_chunks,
    meta_dataset_collator,
)
from tabpfn.finetuning.train_util import (
    get_and_init_optimizer,
    get_checkpoint_path_and_epoch_from_output_dir,
    get_cosine_schedule_with_warmup,
    save_checkpoint,
)
from tabpfn.utils import infer_devices
from tabpfn.validation import ensure_compatible_fit_inputs_sklearn

logger = logging.getLogger(__name__)

if TYPE_CHECKING:
    from tabpfn.constants import XType, YType

# Currently, we only support a batch size of 1 for finetuning.
META_BATCH_SIZE = 1

# Hard limit on the number of samples to use for validation.
# This is used to avoid spending too much time on validation
# and prevent OOM issues for very large datasets.
MAX_VALIDATION_SAMPLES = 50_000


@dataclass
class EvalResult:
    """Container for evaluation results.

    Attributes:
        primary: The primary metric used for early stopping decisions.
        secondary: Additional metrics for logging purposes only.
    """

    primary: float
    secondary: dict[str, float] = field(default_factory=dict)


class FinetunedTabPFNBase(BaseEstimator, ABC):
    """Abstract base class for fine-tuning TabPFN models.

    This class encapsulates the shared fine-tuning logic, allowing you to
    fine-tune TabPFN on a specific dataset using the familiar .fit() and
    .predict() API.

    Args:
        device: The device to run the model on. Defaults to "cuda".
        epochs: The total number of passes through the fine-tuning data.
            Defaults to 30.
        time_limit: Time limit in seconds for fine-tuning.
            If None, no time limit is applied. Defaults to None.
        learning_rate: The learning rate for the AdamW optimizer. A small value
            is crucial for stable fine-tuning. Defaults to 1e-5.
        weight_decay: The weight decay for the AdamW optimizer. Defaults to 0.01.
        validation_split_ratio: Fraction of the original training data reserved
            as a validation set for early stopping and monitoring. Defaults to 0.1.
        n_finetune_ctx_plus_query_samples: The total number of samples per
            meta-dataset during fine-tuning (context plus query) before applying
            the `finetune_ctx_query_split_ratio`. Defaults to 10_000.
        finetune_ctx_query_split_ratio: The proportion of each fine-tuning
            meta-dataset to use as query samples for calculating the loss. The
            remainder is used as context. Defaults to 0.2.
        n_inference_subsample_samples: The total number of subsampled training
            samples per estimator during validation and final inference.
            Defaults to 50_000.
        random_state: Seed for reproducibility of data splitting and model
            initialization. Defaults to 0.
        early_stopping: Whether to use early stopping based on validation
            performance. Defaults to True.
        early_stopping_patience: Number of epochs to wait for improvement before
            early stopping. Defaults to 8.
        min_delta: Minimum change in metric to be considered as an improvement.
            Defaults to 1e-4.
        grad_clip_value: Maximum norm for gradient clipping. If None, gradient
            clipping is disabled. Gradient clipping helps stabilize training by
            preventing exploding gradients. Defaults to 1.0.
        use_lr_scheduler: Whether to use a learning rate scheduler (linear warmup
            with optional cosine decay) during fine-tuning. Defaults to True.
        lr_warmup_only: If True, only performs linear warmup to the base learning
            rate and then keeps it constant. If False, applies cosine decay after
            warmup. Defaults to False.
        n_estimators_finetune: If set, overrides `n_estimators` of the underlying
            estimator only during fine-tuning to control the number of
            estimators (ensemble size) used in the training loop. If None, the
            value from `kwargs` or the estimator default is used.
            Defaults to 2.
        n_estimators_validation: If set, overrides `n_estimators` only for
            validation-time evaluation during fine-tuning (early-stopping /
            monitoring). If None, the value from `kwargs` or the
            estimator default is used. Defaults to 2.
        n_estimators_final_inference: If set, overrides `n_estimators` only for
            the final fitted inference model that is used after fine-tuning. If
            None, the value from `kwargs` or the estimator default is used.
            Defaults to 8.
        use_activation_checkpointing: Whether to use activation checkpointing to
            reduce memory usage. Defaults to True.
        save_checkpoint_interval: Number of epochs between checkpoint saves. This
            only has an effect if `output_dir` is provided during the `fit()` call.
            If None, no intermediate checkpoints are saved. The best model checkpoint
            is always saved regardless of this setting. Defaults to 10.
    """

    def __init__(  # noqa: PLR0913
        self,
        *,
        device: str = "cuda",
        epochs: int = 30,
        time_limit: int | None = None,
        learning_rate: float = 1e-5,
        weight_decay: float = 0.01,
        validation_split_ratio: float = 0.1,
        n_finetune_ctx_plus_query_samples: int = 10_000,
        finetune_ctx_query_split_ratio: float = 0.2,
        n_inference_subsample_samples: int = 50_000,
        random_state: int = 0,
        early_stopping: bool = True,
        early_stopping_patience: int = 8,
        min_delta: float = 1e-4,
        grad_clip_value: float | None = 1.0,
        use_lr_scheduler: bool = True,
        lr_warmup_only: bool = False,
        n_estimators_finetune: int = 2,
        n_estimators_validation: int = 2,
        n_estimators_final_inference: int = 8,
        use_activation_checkpointing: bool = True,
        save_checkpoint_interval: int | None = 10,
    ):
        super().__init__()
        self.device = device
        self.epochs = epochs
        self.time_limit = time_limit
        self.learning_rate = learning_rate
        self.weight_decay = weight_decay
        self.validation_split_ratio = validation_split_ratio
        self.n_finetune_ctx_plus_query_samples = n_finetune_ctx_plus_query_samples
        self.finetune_ctx_query_split_ratio = finetune_ctx_query_split_ratio
        self.n_inference_subsample_samples = n_inference_subsample_samples
        self.random_state = random_state
        self.early_stopping = early_stopping
        self.early_stopping_patience = early_stopping_patience
        self.min_delta = min_delta
        self.grad_clip_value = grad_clip_value
        self.use_lr_scheduler = use_lr_scheduler
        self.lr_warmup_only = lr_warmup_only
        self.n_estimators_finetune = n_estimators_finetune
        self.n_estimators_validation = n_estimators_validation
        self.n_estimators_final_inference = n_estimators_final_inference
        self.use_activation_checkpointing = use_activation_checkpointing
        self.save_checkpoint_interval = save_checkpoint_interval
        self.meta_batch_size = META_BATCH_SIZE

    def _build_estimator_config(
        self,
        base_config: dict[str, Any],
        n_estimators_override: int | None,
    ) -> dict[str, Any]:
        """Return a deep-copy of base_config with an optional n_estimators override."""
        config = copy.deepcopy(base_config)
        if n_estimators_override is not None:
            config["n_estimators"] = n_estimators_override
        return config

    def _build_eval_config(
        self,
        base_config: dict[str, Any],
        n_estimators_override: int | None,
    ) -> dict[str, Any]:
        """Return eval config with n_estimators override and subsample setting."""
        config = self._build_estimator_config(base_config, n_estimators_override)
        existing = dict(config.get("inference_config", {}) or {})
        existing["SUBSAMPLE_SAMPLES"] = self.n_inference_subsample_samples
        config["inference_config"] = existing
        return config

    @property
    @abstractmethod
    def _estimator_kwargs(self) -> dict[str, Any]:
        """Return the task-specific estimator kwargs."""
        ...

    @property
    @abstractmethod
    def _model_type(self) -> Literal["classifier", "regressor"]:
        """Return the model type string ('classifier' or 'regressor')."""
        ...

    @property
    @abstractmethod
    def _metric_name(self) -> str:
        """Return the name of the primary metric for logging."""
        ...

    @abstractmethod
    def _create_estimator(self, config: dict[str, Any]) -> Any:
        """Create and return the underlying TabPFN estimator with the given config."""
        ...

    @abstractmethod
    def _setup_estimator(self) -> None:
        """Perform any task-specific setup after estimator creation."""
        ...

    @abstractmethod
    def _setup_batch(self, batch: ClassifierBatch | RegressorBatch) -> None:
        """Perform any batch-specific setup before the forward pass."""
        ...

    @abstractmethod
    def _should_skip_batch(self, batch: ClassifierBatch | RegressorBatch) -> bool:
        """Check if the batch should be skipped."""
        ...

    @abstractmethod
    def _forward_with_loss(
        self,
        batch: ClassifierBatch | RegressorBatch,
    ) -> torch.Tensor:
        """Perform forward pass and compute loss for the given batch.

        Args:
            batch: The batch tuple from the dataloader.

        Returns:
            The computed loss tensor.
        """
        ...

    @abstractmethod
    def _evaluate_model(
        self,
        eval_config: dict[str, Any],
        X_train: np.ndarray,
        y_train: np.ndarray,
        X_val: np.ndarray,
        y_val: np.ndarray,
    ) -> EvalResult:
        """Evaluate the model on validation data and return metrics.

        Args:
            eval_config: Configuration dictionary for the evaluation estimator.
            X_train: Training input samples.
            y_train: Training target values.
            X_val: Validation input samples.
            y_val: Validation target values.

        Returns:
            EvalResult with primary metric for early stopping and secondary
            metrics for logging.
        """
        ...

    @abstractmethod
    def _is_improvement(self, current: float, best: float) -> bool:
        """Return True if current metric is an improvement over best.

        Args:
            current: The current metric value.
            best: The best metric value seen so far.

        Returns:
            True if current is better than best (accounting for min_delta).
        """
        ...

    @abstractmethod
    def _get_initial_best_metric(self) -> float:
        """Return initial 'best' metric (inf for min, -inf for max)."""
        ...

    @abstractmethod
    def _get_checkpoint_metrics(self, eval_result: EvalResult) -> dict[str, float]:
        """Return the metrics dict to save in checkpoints."""
        ...

    @abstractmethod
    def _log_epoch_evaluation(
        self, epoch: int, eval_result: EvalResult, mean_train_loss: float | None
    ) -> None:
        """Log the evaluation results for the current epoch."""
        ...

    @abstractmethod
    def _setup_inference_model(
        self, final_inference_eval_config: dict[str, Any]
    ) -> None:
        """Set up the final inference model after fine-tuning completes."""
        ...

    @abstractmethod
    def predict(self, X: np.ndarray) -> np.ndarray:
        """Predict target values for X."""
        ...

    def _get_train_val_split(
        self, X: np.ndarray, y: np.ndarray
    ) -> tuple[np.ndarray, np.ndarray, np.ndarray, np.ndarray]:
        """Split data into train/validation sets with task-specific options."""
        n_samples = len(y)
        test_size = int(n_samples * self.validation_split_ratio)

        if test_size > MAX_VALIDATION_SAMPLES:
            warnings.warn(
                f"Validation set size would be {test_size:,} samples "
                f"based on validation_split_ratio="
                f"{self.validation_split_ratio:.2f}, but limiting to "
                f"{MAX_VALIDATION_SAMPLES:,} samples to avoid excessive "
                f"validation time and memory usage.",
                UserWarning,
                stacklevel=3,
            )
            test_size = MAX_VALIDATION_SAMPLES

        # test_size should be greater or equal to the number of classes
        if self._model_type == "classifier":
            n_classes = len(np.unique(y))
            test_size = max(test_size, n_classes)

        return train_test_split(  # type: ignore[return-value]
            X,
            y,
            test_size=test_size,
            random_state=self.random_state,
            stratify=y if self._model_type == "classifier" else None,
        )

    @abstractmethod
    def _get_valid_finetuning_query_size(
        self, *, query_size: int, y_train: np.ndarray | None
    ) -> int:
        """Calculate a valid finetuning query size."""
        ...

    def fit(
        self,
        X: XType,
        y: YType,
        X_val: XType | None = None,
        y_val: YType | None = None,
        output_dir: Path | None = None,
    ) -> FinetunedTabPFNBase:
        """Fine-tune the TabPFN model on the provided training data.

        Args:
            X: The training input samples of shape (n_samples, n_features).
            y: The target values of shape (n_samples,).
            X_val: Optional validation input samples.
            y_val: Optional validation target values.
            output_dir: Directory path for saving checkpoints. If None, no
                checkpointing is performed and progress will be lost if
                training is interrupted.

        Returns:
            The fitted instance itself.
        """
        if output_dir is None:
            warnings.warn(
                "`output_dir` is not set. This means no checkpointing will be done and "
                "all progress will be lost if the training is interrupted.",
                UserWarning,
                stacklevel=2,
            )
        else:
            output_dir.mkdir(parents=True, exist_ok=True)

        return self._fit(X=X, y=y, X_val=X_val, y_val=y_val, output_dir=output_dir)

    def _fit(  # noqa: C901,PLR0912
        self,
        X: XType,
        y: YType,
        X_val: XType | None = None,
        y_val: YType | None = None,
        output_dir: Path | None = None,
    ) -> FinetunedTabPFNBase:
        """Internal implementation of fit that runs the finetuning loop."""
        # Store the original training size for checkpoint naming
        train_size = X.shape[0]
        start_time = time.monotonic()

        _estimator_kwargs = copy.deepcopy(self._estimator_kwargs)
        model_path = _estimator_kwargs.pop("model_path", None)
        inference_config = copy.deepcopy(_estimator_kwargs.get("inference_config", {}))
        base_estimator_config: dict[str, Any] = {
            **_estimator_kwargs,
            "ignore_pretraining_limits": True,
            "device": self.device,
            "random_state": self.random_state,
            "inference_config": inference_config,
        }

        # Config used for the finetuning loop.
        finetuning_estimator_config = self._build_estimator_config(
            base_estimator_config,
            self.n_estimators_finetune,
        )
        if model_path is not None:
            finetuning_estimator_config["model_path"] = model_path

        # Configs used for validation-time evaluation and final inference.
        validation_eval_config = self._build_eval_config(
            base_estimator_config,
            self.n_estimators_validation,
        )
        final_inference_eval_config = self._build_eval_config(
            base_estimator_config,
            self.n_estimators_final_inference,
        )

        eval_devices = infer_devices(self.device)
        validation_eval_config["device"] = eval_devices
        final_inference_eval_config["device"] = eval_devices

        epoch_to_start_from = 0
        checkpoint_path = None
        if output_dir is not None:
            checkpoint_path, epoch_to_start_from = (
                get_checkpoint_path_and_epoch_from_output_dir(
                    output_dir=output_dir,
                    train_size=train_size,
                    get_best=False,
                )
            )
            if checkpoint_path is not None:
                logger.info(
                    f"Restarting training from checkpoint {checkpoint_path} at epoch "
                    f"{epoch_to_start_from}",
                )
                finetuning_estimator_config["model_path"] = checkpoint_path

        self.finetuned_estimator_ = self._create_estimator(finetuning_estimator_config)
        self._setup_estimator()

        X, y, _, _ = ensure_compatible_fit_inputs_sklearn(
            X,
            y,
            estimator=self.finetuned_estimator_,
            ensure_y_numeric=self._model_type == "regressor",
        )

        self.X_ = X
        self.y_ = y
        if X_val is not None and y_val is not None:
            X_train, y_train = X, y
            X_val, y_val, _, _ = ensure_compatible_fit_inputs_sklearn(
                X_val,
                y_val,
                estimator=self.finetuned_estimator_,
                ensure_y_numeric=self._model_type == "regressor",
            )
        else:
            X_train, X_val, y_train, y_val = self._get_train_val_split(X, y)

        # Calculate the context size used during finetuning.
        n_finetune_ctx_plus_query_samples = min(
            self.n_finetune_ctx_plus_query_samples,
            len(y_train),
        )

        self.finetuned_estimator_._initialize_model_variables()
        self.finetuned_estimator_.model_.to(self.device)

        if self.use_activation_checkpointing:
            self.finetuned_estimator_.model_.recompute_layer = True  # type: ignore

        optimizer = get_and_init_optimizer(
            model_parameters=self.finetuned_estimator_.model_.parameters(),  # type: ignore
            learning_rate=self.learning_rate,
            weight_decay=self.weight_decay,
            checkpoint_path=checkpoint_path,
            device=self.device,
        )

        use_amp = self.device.startswith("cuda") and torch.cuda.is_available()
        scaler = GradScaler() if use_amp else None  # type: ignore

        logger.info("--- 🚀 Eval default model ---")
        eval_result = self._evaluate_model(
            validation_eval_config,
            X_train,  # pyright: ignore[reportArgumentType]
            y_train,  # pyright: ignore[reportArgumentType]
            X_val,  # pyright: ignore[reportArgumentType]
            y_val,  # pyright: ignore[reportArgumentType]
        )
        self._log_epoch_evaluation(-1, eval_result, mean_train_loss=None)
        best_metric: float = eval_result.primary

        logger.info("--- 🚀 Starting Fine-tuning ---")
        patience_counter = 0
        best_model = None

        scheduler: LambdaLR | None = None

        finetuning_query_size = self._get_valid_finetuning_query_size(
            query_size=int(
                n_finetune_ctx_plus_query_samples * self.finetune_ctx_query_split_ratio
            ),
            y_train=y_train,
        )
        for epoch in range(epoch_to_start_from, self.epochs):
            # Per-epoch aggregates for cleaner learning curves.
            epoch_loss_sum = 0.0
            epoch_batches = 0

            # Regenerate datasets each epoch with a different random_state
            training_splitter = partial(
                train_test_split,
                test_size=finetuning_query_size,
                random_state=self.random_state + epoch,
            )

            training_datasets = get_preprocessed_dataset_chunks(
                calling_instance=self.finetuned_estimator_,
                X_raw=X_train,
                y_raw=y_train,
                split_fn=training_splitter,
                max_data_size=n_finetune_ctx_plus_query_samples,
                model_type=self._model_type,
                equal_split_size=False,
                seed=self.random_state + epoch,
            )

            dataloader_generator = torch.Generator().manual_seed(
                self.random_state + epoch
            )
            finetuning_dataloader = DataLoader(
                training_datasets,
                batch_size=self.meta_batch_size,
                collate_fn=meta_dataset_collator,
                shuffle=True,
                generator=dataloader_generator,
            )

            # Instantiate the LR scheduler only once
            if self.use_lr_scheduler and scheduler is None:
                steps_per_epoch = len(finetuning_dataloader)
                if steps_per_epoch == 0:
                    logger.warning(
                        "No training batches available; ending training early.",
                    )
                    break

                total_steps = steps_per_epoch * self.epochs
                warmup_steps = int(total_steps * 0.1)

                lrate_schedule_fn = get_cosine_schedule_with_warmup(
                    total_steps=total_steps,
                    warmup_steps=warmup_steps,
                    warmup_only=self.lr_warmup_only,
                )
                scheduler = LambdaLR(optimizer, lr_lambda=lrate_schedule_fn)

                logger.info(
                    "Using LambdaLR %s schedule: total_steps=%d, warmup_steps=%d",
                    "warmup-only (constant LR after warmup)"
                    if self.lr_warmup_only
                    else "warmup+cosine",
                    total_steps,
                    warmup_steps,
                )

            progress_bar = tqdm(
                finetuning_dataloader,
                desc=f"Finetuning Epoch {epoch + 1}/{self.epochs}",
            )
            for batch in progress_bar:
                optimizer.zero_grad()

                if self._should_skip_batch(batch):
                    continue

                self._setup_batch(batch)

                self.finetuned_estimator_.fit_from_preprocessed(
                    batch.X_context,
                    batch.y_context,
                    batch.cat_indices,
                    batch.configs,
                )

                use_scaler = use_amp and scaler is not None

                with autocast(enabled=use_scaler), sdpa_kernel_context():  # type: ignore
                    loss = self._forward_with_loss(batch)

                if use_scaler:
                    with sdpa_kernel_context():
                        scaler.scale(loss).backward()  # type: ignore
                    scaler.unscale_(optimizer)  # type: ignore

                    if self.grad_clip_value is not None:
                        clip_grad_norm_(
                            self.finetuned_estimator_.model_.parameters(),  # type: ignore
                            self.grad_clip_value,
                        )

                    scaler.step(optimizer)  # type: ignore
                    scaler.update()  # type: ignore
                else:
                    with sdpa_kernel_context():
                        loss.backward()

                    if self.grad_clip_value is not None:
                        clip_grad_norm_(
                            self.finetuned_estimator_.model_.parameters(),  # type: ignore
                            self.grad_clip_value,
                        )

                    optimizer.step()

                if scheduler is not None:
                    scheduler.step()

                loss_scalar = float(loss.detach().item())

                epoch_loss_sum += loss_scalar
                epoch_batches += 1

                progress_bar.set_postfix(
                    loss=f"{loss_scalar:.4f}",
                )

            mean_train_loss = (
                epoch_loss_sum / epoch_batches if epoch_batches > 0 else None
            )

            eval_result = self._evaluate_model(
                validation_eval_config,
                X_train,  # pyright: ignore[reportArgumentType]
                y_train,  # pyright: ignore[reportArgumentType]
                X_val,  # pyright: ignore[reportArgumentType]
                y_val,  # pyright: ignore[reportArgumentType]
            )

            self._log_epoch_evaluation(epoch, eval_result, mean_train_loss)

            primary_metric = eval_result.primary

            if output_dir is not None and not np.isnan(primary_metric):
                save_interval_checkpoint = (
                    self.save_checkpoint_interval is not None
                    and (epoch + 1) % self.save_checkpoint_interval == 0
                )

                is_best = self._is_improvement(primary_metric, best_metric)

                if save_interval_checkpoint or is_best:
                    save_checkpoint(
                        estimator=self.finetuned_estimator_,
                        output_dir=output_dir,
                        epoch=epoch + 1,
                        optimizer=optimizer,
                        metrics=self._get_checkpoint_metrics(eval_result),
                        train_size=train_size,
                        is_best=is_best,
                        save_interval_checkpoint=save_interval_checkpoint,
                    )

            if self.early_stopping and not np.isnan(primary_metric):
                if self._is_improvement(primary_metric, best_metric):
                    best_metric = primary_metric
                    patience_counter = 0
                    best_model = copy.deepcopy(self.finetuned_estimator_)
                else:
                    patience_counter += 1
                    logger.info(
                        "⚠️  No improvement for %s epochs. Best %s: %.4f",
                        patience_counter,
                        self._metric_name,
                        best_metric,
                    )

                if patience_counter >= self.early_stopping_patience:
                    logger.info(
                        "🛑 Early stopping triggered. Best %s: %.4f",
                        self._metric_name,
                        best_metric,
                    )
                    if best_model is not None:
                        self.finetuned_estimator_ = best_model
                    break

            if self.time_limit is not None:
                elapsed_time = time.monotonic() - start_time
                if elapsed_time > self.time_limit:
                    logger.info(
                        "🛑 Time limit of %d seconds reached. Stopping training.",
                        self.time_limit,
                    )
                    break

                n_epochs_run = epoch + 1 - epoch_to_start_from
                if elapsed_time + (elapsed_time / n_epochs_run) > self.time_limit:
                    logger.info(
                        "🛑 Not enough time remaining for another epoch. Stopping "
                        "training.",
                    )
                    break

        if self.early_stopping and best_model is not None:
            self.finetuned_estimator_ = best_model

        logger.info("--- ✅ Fine-tuning Finished ---")

        self._setup_inference_model(final_inference_eval_config)

        self.is_fitted_ = True
        return self



"""Torch operations for preprocessing with NaN handling."""

from __future__ import annotations

import torch


def torch_nansum(
    x: torch.Tensor,
    axis: int | tuple[int, ...] | None = None,
    *,
    keepdim: bool = False,
    dtype: torch.dtype | None = None,
) -> torch.Tensor:
    """Compute the sum of a tensor, treating NaNs as zero.

    Args:
        x: The input tensor.
        axis: The dimension or dimensions to reduce.
        keepdim: Whether the output tensor has `dims` retained or not.
        dtype: The desired data type of the returned tensor.

    Returns:
        The sum of the tensor with NaNs treated as zero.
    """
    nan_mask = torch.isnan(x)
    masked_input = torch.where(
        nan_mask,
        torch.tensor(0.0, device=x.device, dtype=x.dtype),
        x,
    )
    return torch.sum(masked_input, dim=axis, keepdim=keepdim, dtype=dtype)


def torch_nanmean(
    x: torch.Tensor,
    axis: int = 0,
    *,
    include_inf: bool = False,
) -> torch.Tensor:
    """Compute the mean of a tensor over a given dimension, ignoring NaNs.

    Args:
        x: The input tensor.
        axis: The dimension to reduce.
        include_inf: If True, treat infinity as NaN for the purpose of the calculation.

    Returns:
        The mean of the input tensor, ignoring NaNs.
    """
    nan_mask = torch.isnan(x)
    if include_inf:
        nan_mask = torch.logical_or(nan_mask, torch.isinf(x))

    num_valid = torch.where(
        nan_mask,
        torch.zeros_like(x),
        torch.ones_like(x),
    ).sum(dim=axis)
    value_sum = torch.where(nan_mask, torch.zeros_like(x), x).sum(dim=axis)

    return value_sum / num_valid.clamp(min=1.0)


def torch_nanstd(x: torch.Tensor, axis: int = 0) -> torch.Tensor:
    """Compute standard deviation of a tensor over a given dimension, ignoring NaNs.

    Args:
        x: The input tensor.
        axis: The dimension to reduce.

    Returns:
        The standard deviation of the input tensor, ignoring NaNs.
    """
    nan_mask = torch.isnan(x)
    num_valid = torch.where(
        nan_mask,
        torch.zeros_like(x),
        torch.ones_like(x),
    ).sum(dim=axis)
    value_sum = torch.where(nan_mask, torch.zeros_like(x), x).sum(dim=axis)

    mean = value_sum / num_valid.clamp(min=1.0)

    # Broadcast mean back to original shape for subtraction
    mean_broadcast = mean.unsqueeze(axis).expand_as(x)

    # Compute sum of squared differences, ignoring NaNs
    sq_diff = torch.where(
        nan_mask,
        torch.zeros_like(x),
        torch.square(x - mean_broadcast),
    ).sum(dim=axis)

    # Use correction (N-1) to match sklearn's behavior
    variance = sq_diff / (num_valid - 1).clamp(min=1.0)

    return torch.sqrt(variance)



"""Tests for TabPFN regressor finetuning functionality.

This module contains regressor-specific tests for:
- The FinetunedTabPFNRegressor wrapper class (.fit() / .predict()).
- Regression checkpoint metric fields (e.g. storing 'mse').

We intentionally avoid duplicating tests that primarily exercise common logic in
`FinetunedTabPFNBase`, since those are covered by the classifier finetuning tests.
"""

from __future__ import annotations

from collections.abc import Callable
from pathlib import Path
from unittest import mock

import numpy as np
import pytest
import torch
from sklearn.datasets import make_regression
from sklearn.model_selection import train_test_split
from torch.utils.data import DataLoader

from tabpfn.architectures.base.bar_distribution import BarDistribution
from tabpfn.finetuning.data_util import (
    RegressorBatch,
    get_preprocessed_dataset_chunks,
    meta_dataset_collator,
)
from tabpfn.finetuning.finetuned_regressor import (
    FinetunedTabPFNRegressor,
    _compute_regression_loss,
)
from tabpfn.preprocessing import RegressorEnsembleConfig
from tabpfn.regressor import TabPFNRegressor

from .utils import (
    get_pytest_devices,
    get_pytest_devices_with_mps_marked_slow,
    mark_mps_configs_as_slow,
)

rng = np.random.default_rng(42)

devices = get_pytest_devices()


def create_mock_architecture_forward_regression() -> Callable[..., torch.Tensor]:
    """Return a side_effect for mocking the internal Architecture forward in regression.

    The Architecture.forward method signature is:
    forward(x, y, *, only_return_standard_out=True, categorical_inds=None)

    Where:
    - x has shape (train+test rows, batch_size, num_features)
    - y has shape (train rows, batch_size) or (train rows, batch_size, 1)
    - returns shape (test rows, batch_size, n_out), with n_out determined by the model.
    """

    def mock_forward(
        self: torch.nn.Module,
        x: torch.Tensor | dict[str, torch.Tensor],
        y: torch.Tensor | dict[str, torch.Tensor] | None,
        **_kwargs: bool,
    ) -> torch.Tensor:
        """Mock forward pass that returns random logits with the correct shape."""
        if isinstance(x, dict):
            x = x["main"]

        if y is not None:
            y_tensor = y["main"] if isinstance(y, dict) else y
            num_train_rows = y_tensor.shape[0]
        else:
            num_train_rows = 0

        total_rows = x.shape[0]
        batch_size = x.shape[1]
        num_test_rows = total_rows - num_train_rows

        # Touch a model parameter so gradients flow during backward pass.
        # This mirrors the classifier tests and avoids GradScaler issues on CUDA.
        first_param = next(self.parameters())
        param_contribution = 0.0 * first_param.sum()

        n_out = int(getattr(self, "n_out", 1))
        return (
            torch.randn(
                num_test_rows,
                batch_size,
                n_out,
                requires_grad=True,
                device=x.device,
            )
            + param_contribution
        )

    return mock_forward


@pytest.fixture(scope="module")
def synthetic_regression_data() -> tuple[np.ndarray, np.ndarray]:
    """Generate synthetic regression data for testing."""
    result = make_regression(
        n_samples=120,
        n_features=6,
        n_informative=4,
        noise=0.1,
        random_state=42,
        coef=False,
    )
    X = np.asarray(result[0], dtype=np.float32)
    y = np.asarray(result[1], dtype=np.float32)
    return X, y


@pytest.fixture(params=devices)
def regressor_instance(request: pytest.FixtureRequest) -> TabPFNRegressor:
    """Provide a basic regressor instance, parameterized by device."""
    device = request.param
    if device == "cuda" and not torch.cuda.is_available():
        pytest.skip("CUDA device requested but not available.")
    return TabPFNRegressor(
        n_estimators=2,
        device=device,
        random_state=42,
        inference_precision=torch.float32,
        fit_mode="batched",
        differentiable_input=False,
    )


@pytest.fixture
def variable_synthetic_regression_dataset_collection() -> list[
    tuple[np.ndarray, np.ndarray]
]:
    """Create a small collection of synthetic regression datasets with varying sizes."""
    datasets = []
    dataset_sizes = [10, 20, 30]
    num_features = 3
    for num_samples in dataset_sizes:
        X = rng.normal(size=(num_samples, num_features)).astype(np.float32)
        y = rng.normal(size=(num_samples,)).astype(np.float32)
        datasets.append((X, y))
    return datasets


@pytest.mark.parametrize(
    ("device", "early_stopping", "use_lr_scheduler"),
    mark_mps_configs_as_slow(
        (device, early_stopping, use_lr_scheduler)
        for device in devices
        for early_stopping in [True, False]
        for use_lr_scheduler in [True, False]
    ),
)
def test__finetuned_tabpfn_regressor__fit_and_predict(
    device: str,
    early_stopping: bool,
    use_lr_scheduler: bool,
    synthetic_regression_data: tuple[np.ndarray, np.ndarray],
) -> None:
    """Test FinetunedTabPFNRegressor fit/predict with a mocked forward pass."""
    if device == "cuda" and not torch.cuda.is_available():
        pytest.skip("CUDA device requested but not available.")

    X, y = synthetic_regression_data
    X_train, X_test, y_train, _y_test = train_test_split(
        X, y, test_size=0.3, random_state=42
    )
    X_train = np.asarray(X_train)
    X_test = np.asarray(X_test)
    y_train = np.asarray(y_train)

    epochs = 4 if early_stopping else 2
    finetuned_reg = FinetunedTabPFNRegressor(
        device=device,
        epochs=epochs,
        learning_rate=1e-4,
        validation_split_ratio=0.2,
        n_finetune_ctx_plus_query_samples=60,
        finetune_ctx_query_split_ratio=0.2,
        n_inference_subsample_samples=120,
        random_state=42,
        early_stopping=early_stopping,
        early_stopping_patience=2,
        n_estimators_finetune=1,
        n_estimators_validation=1,
        n_estimators_final_inference=1,
        use_lr_scheduler=use_lr_scheduler,
        lr_warmup_only=False,
    )

    mock_forward = create_mock_architecture_forward_regression()
    with mock.patch(
        "tabpfn.architectures.base.transformer.PerFeatureTransformer.forward",
        autospec=True,
        side_effect=mock_forward,
    ):
        finetuned_reg.fit(X_train, y_train)

    assert finetuned_reg.is_fitted_
    assert hasattr(finetuned_reg, "finetuned_estimator_")
    assert hasattr(finetuned_reg, "finetuned_inference_regressor_")

    predictions = finetuned_reg.predict(X_test)
    assert predictions.shape == (X_test.shape[0],)
    assert np.isfinite(predictions).all()


@pytest.mark.parametrize("device", get_pytest_devices_with_mps_marked_slow())
def test__regressor_checkpoint_contains_mse_metric(
    device: str,
    tmp_path: Path,
    synthetic_regression_data: tuple[np.ndarray, np.ndarray],
) -> None:
    """Ensure regressor checkpoints store regression metrics (mse).

    This also checks that classifier-only metric fields are not stored.
    """
    if device == "cuda" and not torch.cuda.is_available():
        pytest.skip("CUDA device requested but not available.")

    X, y = synthetic_regression_data
    X_train, _X_test, y_train, _y_test = train_test_split(
        X, y, test_size=0.3, random_state=42
    )
    X_train = np.asarray(X_train)
    y_train = np.asarray(y_train)
    output_folder = tmp_path / "checkpoints_regressor"

    finetuned_reg = FinetunedTabPFNRegressor(
        device=device,
        epochs=2,
        learning_rate=1e-4,
        validation_split_ratio=0.2,
        n_finetune_ctx_plus_query_samples=60,
        finetune_ctx_query_split_ratio=0.2,
        n_inference_subsample_samples=120,
        random_state=42,
        early_stopping=False,
        use_lr_scheduler=False,
        n_estimators_finetune=1,
        n_estimators_validation=1,
        n_estimators_final_inference=1,
        save_checkpoint_interval=1,
    )

    mock_forward = create_mock_architecture_forward_regression()
    with mock.patch(
        "tabpfn.architectures.base.transformer.PerFeatureTransformer.forward",
        autospec=True,
        side_effect=mock_forward,
    ):
        finetuned_reg.fit(X_train, y_train, output_dir=output_folder)

    best_checkpoint_candidates = list(output_folder.glob("checkpoint_*_best.pth"))
    assert len(best_checkpoint_candidates) == 1, "Expected exactly one best checkpoint."
    best_checkpoint_path = best_checkpoint_candidates[0]

    best_checkpoint = torch.load(best_checkpoint_path, weights_only=False)
    assert "state_dict" in best_checkpoint
    assert "config" in best_checkpoint
    assert "optimizer" in best_checkpoint
    assert "epoch" in best_checkpoint
    assert "mse" in best_checkpoint
    assert "roc_auc" not in best_checkpoint
    assert "log_loss" not in best_checkpoint


def test__compute_regression_loss__correct_mse_and_mae_with_nan_targets() -> None:
    """Ensure NaN targets are ignored for auxiliary regression losses."""
    borders = torch.linspace(-2.0, 2.0, steps=9, dtype=torch.float32)
    bardist_loss_fn = BarDistribution(borders=borders, ignore_nan_targets=True)

    # Deterministic setup: zero logits -> uniform bar probabilities.
    # For symmetric borders, this yields an exact mean prediction of 0.0.
    logits_BQL = torch.zeros((1, 4, bardist_loss_fn.num_bars), dtype=torch.float32)
    predictions_mean_BQ = bardist_loss_fn.mean(logits_BQL)
    assert torch.equal(predictions_mean_BQ, torch.zeros_like(predictions_mean_BQ))

    targets_BQ = torch.tensor([[0.0, 0.5, float("nan"), -0.5]], dtype=torch.float32)
    assert torch.isnan(targets_BQ[0, 2]).item()

    mse_mae_only_loss = _compute_regression_loss(
        logits_BQL=logits_BQL,
        targets_BQ=targets_BQ,
        bardist_loss_fn=bardist_loss_fn,
        ce_loss_weight=0.0,
        mse_loss_weight=1.0,
        mse_loss_clip=None,
        mae_loss_weight=1.0,
        mae_loss_clip=None,
    )
    assert torch.isfinite(mse_mae_only_loss).item()
    # MSE terms (valid only): (0.0 - 0.0)^2 + (0.0 - 0.5)^2 + (0.0 - (-0.5))^2 = 0.5
    # Mean is over all 4 positions (NaN position contributes 0.0): 0.5 / 4 = 0.125
    expected_mse = 0.125
    # MAE terms (valid only): |0.0 - 0.0| + |0.0 - 0.5| + |0.0 - (-0.5)| = 1.0
    # Mean over all 4 positions: 1.0 / 4 = 0.25
    expected_mae = 0.25
    expected_total = expected_mse + expected_mae
    assert mse_mae_only_loss.item() == expected_total


def test__compute_regression_loss__masks_nan_targets() -> None:
    borders = torch.linspace(-2.0, 2.0, steps=9, dtype=torch.float32)
    bardist_loss_fn = BarDistribution(borders=borders, ignore_nan_targets=True)
    logits_BQL = torch.zeros((1, 4, bardist_loss_fn.num_bars), dtype=torch.float32)

    rps_only_loss = _compute_regression_loss(
        logits_BQL=logits_BQL,
        targets_BQ=torch.full((1, 4), float("nan"), dtype=torch.float32),
        bardist_loss_fn=bardist_loss_fn,
        ce_loss_weight=0.0,
        crps_loss_weight=1.0,
        mse_loss_weight=1.0,
        mse_loss_clip=None,
        mae_loss_weight=1.0,
        mae_loss_clip=None,
    )
    assert torch.isfinite(rps_only_loss).item()
    assert rps_only_loss.item() == 0.0

    rls_only_loss = _compute_regression_loss(
        logits_BQL=logits_BQL,
        targets_BQ=torch.full((1, 4), float("nan"), dtype=torch.float32),
        bardist_loss_fn=bardist_loss_fn,
        ce_loss_weight=0.0,
        crls_loss_weight=1.0,
        mse_loss_weight=1.0,
        mse_loss_clip=None,
        mae_loss_weight=1.0,
        mae_loss_clip=None,
    )
    assert torch.isfinite(rls_only_loss).item()
    assert rls_only_loss.item() == 0.0


def test__compute_regression_loss__rps_vs_rls_matches_expected_value() -> None:
    """Check RPS (squared) vs RLS (log) ranked probability loss.

    Uses a deterministic 2-bin toy example so expected values are easy to verify.
    """
    borders = torch.tensor([0.0, 1.0, 2.0])
    bardist_loss_fn = BarDistribution(borders=borders, ignore_nan_targets=True)

    # Two bins. Desired probs: [0.25, 0.75] -> pred CDF: [0.25, 1.0]
    probs_L = torch.tensor([0.25, 0.75])
    logits_L = probs_L.log()
    logits_BQL = logits_L.view(1, 1, -1)

    # Target y in bin 1 -> target CDF: [0.0, 1.0]
    targets_BQ = torch.tensor([[1.2]])

    rps_loss = _compute_regression_loss(
        logits_BQL=logits_BQL,
        targets_BQ=targets_BQ,
        bardist_loss_fn=bardist_loss_fn,
        ce_loss_weight=0.0,
        crps_loss_weight=1.0,
        mse_loss_weight=0.0,
        mse_loss_clip=None,
        mae_loss_weight=0.0,
        mae_loss_clip=None,
    )

    rls_loss = _compute_regression_loss(
        logits_BQL=logits_BQL,
        targets_BQ=targets_BQ,
        bardist_loss_fn=bardist_loss_fn,
        ce_loss_weight=0.0,
        crls_loss_weight=1.0,
        mse_loss_weight=0.0,
        mse_loss_clip=None,
        mae_loss_weight=0.0,
        mae_loss_clip=None,
    )

    # second term is zero but is added for clarity
    expected_rps = (0.25 - 0) ** 2 + (1 - 1) ** 2

    # second term is zero but just added here for clarity
    expected_rls = -np.log(abs(0.25 + 0 - 1)) - np.log(abs(1 + 1 - 1))

    assert rls_loss.item() == pytest.approx(expected_rls)
    assert rps_loss.item() == pytest.approx(expected_rps)

    combined_loss = _compute_regression_loss(
        logits_BQL=logits_BQL,
        targets_BQ=targets_BQ,
        bardist_loss_fn=bardist_loss_fn,
        ce_loss_weight=0.0,
        crps_loss_weight=1.0,
        crls_loss_weight=1.0,
        mse_loss_weight=0.0,
        mse_loss_clip=None,
        mae_loss_weight=0.0,
        mae_loss_clip=None,
    )
    assert combined_loss.item() == pytest.approx(expected_rps + expected_rls)


def test_regressor_dataset_and_collator_batches_type(
    variable_synthetic_regression_dataset_collection: list[
        tuple[np.ndarray, np.ndarray]
    ],
    regressor_instance: TabPFNRegressor,
) -> None:
    """Test that dataset and collator produce correctly-typed RegressorBatch objects."""
    X_list = [X for X, _ in variable_synthetic_regression_dataset_collection]
    y_list = [y for _, y in variable_synthetic_regression_dataset_collection]
    dataset_collection = get_preprocessed_dataset_chunks(
        regressor_instance,
        X_list,
        y_list,
        train_test_split,
        100,
        model_type="regressor",
        equal_split_size=True,
        seed=42,
    )

    dl = DataLoader(
        dataset_collection,
        batch_size=1,
        collate_fn=meta_dataset_collator,
    )
    for batch in dl:
        assert isinstance(batch, RegressorBatch)
        for est_tensor in batch.X_context:
            assert isinstance(est_tensor, torch.Tensor)
            assert est_tensor.shape[0] == 1
        for est_tensor in batch.y_context:
            assert isinstance(est_tensor, torch.Tensor)
            assert est_tensor.shape[0] == 1

        assert isinstance(batch.cat_indices, list)
        for conf in batch.configs:
            for c in conf:
                assert isinstance(c, RegressorEnsembleConfig)

        assert isinstance(batch.X_query_raw, torch.Tensor)
        assert isinstance(batch.y_query_raw, torch.Tensor)
        assert batch.X_query_raw.shape[0] == 1
        assert batch.y_query_raw.shape[0] == 1
        assert batch.y_query.shape[0] == 1
        break



"""Custom implementations of PyTorch functions. Needed for ONNX export."""

from __future__ import annotations

from typing import Literal, overload

import numpy as np
import torch


def torch_nansum(
    x: torch.Tensor,
    axis: int | tuple[int, ...] | None = None,
    *,
    keepdim: bool = False,
    dtype: torch.dtype | None = None,
) -> torch.Tensor:
    """Computes the sum of a tensor, treating NaNs as zero.

    Args:
        x: The input tensor.
        axis: The dimension or dimensions to reduce.
        keepdim: Whether the output tensor has `axis` retained or not.
        dtype: The desired data type of the returned tensor.

    Returns:
        The sum of the tensor with NaNs treated as zero.
    """
    nan_mask = torch.isnan(x)
    masked_input = torch.where(
        nan_mask,
        torch.tensor(0.0, device=x.device, dtype=x.dtype),
        x,
    )
    return torch.sum(masked_input, axis=axis, keepdim=keepdim, dtype=dtype)  # type: ignore


@overload
def torch_nanmean(
    x: torch.Tensor,
    axis: int = 0,
    *,
    return_nanshare: Literal[False] = False,
    include_inf: bool = False,
) -> torch.Tensor: ...


@overload
def torch_nanmean(
    x: torch.Tensor,
    axis: int = 0,
    *,
    return_nanshare: Literal[True],
    include_inf: bool = False,
) -> tuple[torch.Tensor, torch.Tensor]: ...


def torch_nanmean(
    x: torch.Tensor,
    axis: int = 0,
    *,
    return_nanshare: bool = False,
    include_inf: bool = False,
) -> torch.Tensor | tuple[torch.Tensor, torch.Tensor]:
    """Computes the mean of a tensor over a given dimension, ignoring NaNs.

    Designed for stability: If all inputs are NaN, the mean will be 0.0.

    Args:
        x: The input tensor.
        axis: The dimension to reduce.
        return_nanshare: If True, also return the proportion of NaNs.
        include_inf: If True, treat infinity as NaN for the purpose of the calculation.

    Returns:
        The mean of the input tensor, ignoring NaNs. If `return_nanshare` is True,
        returns a tuple containing the mean and the share of NaNs.
    """
    nan_mask = torch.isnan(x)
    if include_inf:
        nan_mask = torch.logical_or(nan_mask, torch.isinf(x))

    num = torch.where(nan_mask, torch.full_like(x, 0), torch.full_like(x, 1)).sum(  # type: ignore
        dim=axis,
    )
    value = torch.where(nan_mask, torch.full_like(x, 0), x).sum(axis=axis)  # type: ignore
    if return_nanshare:
        return value / num, 1.0 - (num / x.shape[axis])
    return value / num.clip(min=1.0)


def torch_nanstd(x: torch.Tensor, axis: int = 0) -> torch.Tensor:
    """Computes standard deviation of a tensor over a given dimension, ignoring NaNs.

    This implementation is designed for stability. It clips the denominator `(num - 1)`
    at a minimum of 1. This prevents division-by-zero errors that would produce `NaN`
    or `inf` when calculating the standard deviation of a single data point.

    Args:
        x: The input tensor.
        axis: The dimension to reduce.

    Returns:
        The standard deviation of the input tensor, ignoring NaNs.
    """
    num = torch.where(torch.isnan(x), torch.full_like(x, 0), torch.full_like(x, 1)).sum(  # type: ignore
        axis=axis,
    )
    value = torch.where(torch.isnan(x), torch.full_like(x, 0), x).sum(axis=axis)  # type: ignore
    mean = value / num.clip(min=1.0)
    mean_broadcast = torch.repeat_interleave(
        mean.unsqueeze(axis),
        x.shape[axis],
        dim=axis,
    )
    # Clip the denominator to avoid division by zero when num=1
    var = torch_nansum(torch.square(mean_broadcast - x), axis=axis) / (num - 1).clip(
        min=1.0
    )
    return torch.sqrt(var)


@overload
def normalize_data(
    data: torch.Tensor,
    *,
    normalize_positions: int = -1,
    return_scaling: Literal[False] = False,
    clip: bool = True,
    std_only: bool = False,
    mean: torch.Tensor | None = None,
    std: torch.Tensor | None = None,
) -> torch.Tensor: ...


@overload
def normalize_data(
    data: torch.Tensor,
    *,
    normalize_positions: int = -1,
    return_scaling: Literal[True],
    clip: bool = True,
    std_only: bool = False,
    mean: torch.Tensor | None = None,
    std: torch.Tensor | None = None,
) -> tuple[torch.Tensor, tuple[torch.Tensor, torch.Tensor]]: ...


def normalize_data(
    data: torch.Tensor,
    *,
    normalize_positions: int = -1,
    return_scaling: bool = False,
    clip: bool = True,
    std_only: bool = False,
    mean: torch.Tensor | None = None,
    std: torch.Tensor | None = None,
) -> torch.Tensor | tuple[torch.Tensor, tuple[torch.Tensor, torch.Tensor]]:
    """Normalize data to mean 0 and std 1 with high numerical stability.

    This function is designed to be robust against several edge cases:
    1.  **Constant Features**: If a feature is constant, its standard deviation (`std`)
        will be 0. This is handled by replacing `std=0` with `1` to prevent `0/0`
        division, effectively mapping constant features to a normalized value of 0.
    2.  **Single-Sample Normalization**: If the normalization is based on a single
        data point, `std` is explicitly set to `1` to prevent undefined behavior.
    3.  **Low-Precision Dtypes**: During the final division, a small epsilon (`1e-16`)
        is added to the denominator. This prevents division by a near-zero `std`,
        which could cause the value to overflow to infinity (`inf`), especially when
        using low-precision dtypes like `torch.float16`.
    4.  **Autograd Compatibility**: All inplace operations (`x[:] = ...`) have been
        replaced with their functional equivalents (`torch.where`, `torch.ones_like`)
        to ensure the computation graph is not corrupted, allowing gradients to be
        computed correctly.

    Args:
        data: The data to normalize. (T, B, H)
        normalize_positions: If > 0, only use the first `normalize_positions`
            positions for normalization.
        return_scaling: If True, return the scaling parameters as well (mean, std).
        std_only: If True, only divide by std.
        clip: If True, clip the data to [-100, 100].
        mean: If given, use this value instead of computing it.
        std: If given, use this value instead of computing it.

    Returns:
        The normalized data tensor, or a tuple containing the data and scaling factors.
    """
    assert (mean is None) == (std is None), (
        "Either both or none of mean and std must be given"
    )
    if mean is None:
        if normalize_positions is not None and normalize_positions > 0:
            mean = torch_nanmean(data[:normalize_positions], axis=0)  # type: ignore
            std = torch_nanstd(data[:normalize_positions], axis=0)
        else:
            mean = torch_nanmean(data, axis=0)  # type: ignore
            std = torch_nanstd(data, axis=0)

        # Inplace assignments with functional equivalents to support autograd.
        std = torch.where(std == 0, torch.ones_like(std), std)

        if len(data) == 1 or normalize_positions == 1:
            std = torch.ones_like(std)

        if std_only:
            mean = torch.zeros_like(mean)

    assert mean is not None
    assert std is not None

    # Add epsilon for numerical stability
    data = (data - mean) / (std + 1e-16)

    if clip:
        data = torch.clip(data, min=-100, max=100)

    if return_scaling:
        return data, (mean, std)
    return data


def select_features(x: torch.Tensor, sel: torch.Tensor) -> torch.Tensor:
    """Select features from the input tensor based on the selection mask,
    and arrange them contiguously in the last dimension.
    If batch size is bigger than 1, we pad the features with zeros to make the number of
    features fixed.

    Args:
        x: The input tensor of shape (sequence_length, batch_size, total_features)
        sel: The boolean selection mask indicating which features to keep of shape
        (batch_size, total_features)

    Returns:
        The tensor with selected features.
        The shape is (sequence_length, batch_size, number_of_selected_features) if
        batch_size is 1.
        The shape is (sequence_length, batch_size, total_features) if batch_size is
        greater than 1.
    """
    B, total_features = sel.shape

    # Do nothing if we need to select all of the features
    if torch.all(sel):
        return x

    # If B == 1, we don't need to append zeros, as the number of features don't need to
    # be fixed.
    if B == 1:
        return x[:, :, sel[0]]

    new_x = x.detach().clone()

    # For each batch, compute the number of selected features.
    sel_counts = sel.sum(dim=-1)  # shape: (B,)

    for b in range(B):
        s = int(sel_counts[b])
        if s != total_features:
            if s > 0:
                new_x[:, b, :s] = x[:, b, sel[b]]
            new_x[:, b, s:] = 0

    return new_x


def remove_outliers(
    X: torch.Tensor,
    n_sigma: float = 4,
    normalize_positions: int = -1,
    lower: None | torch.Tensor = None,
    upper: None | torch.Tensor = None,
) -> tuple[torch.Tensor, tuple[torch.Tensor, torch.Tensor]]:
    # Expects T, B, H
    assert (lower is None) == (upper is None), "Either both or none of lower and upper"
    assert len(X.shape) == 3, "X must be T,B,H"
    # for b in range(X.shape[1]):
    # for col in range(X.shape[2]):
    if lower is None:
        data = X if normalize_positions == -1 else X[:normalize_positions]
        data_clean = data[:].clone()
        data_mean, data_std = torch_nanmean(data, axis=0), torch_nanstd(data, axis=0)
        cut_off = data_std * n_sigma
        lower, upper = data_mean - cut_off, data_mean + cut_off

        data_clean[torch.logical_or(data_clean > upper, data_clean < lower)] = np.nan
        data_mean, data_std = (
            torch_nanmean(data_clean, axis=0),
            torch_nanstd(data_clean, axis=0),
        )
        cut_off = data_std * n_sigma
        lower, upper = data_mean - cut_off, data_mean + cut_off

    assert lower is not None
    assert upper is not None

    X = torch.maximum(-torch.log(1 + torch.abs(X)) + lower, X)
    X = torch.minimum(torch.log(1 + torch.abs(X)) + upper, X)
    return X, (lower, upper)



"""A collection of random utilities for the TabPFN models."""

#  Copyright (c) Prior Labs GmbH 2025.

from __future__ import annotations

import contextlib
import os
from collections.abc import Sequence
from typing import TYPE_CHECKING, Literal, Union

import numpy as np
import numpy.typing as npt
import torch
from sklearn.base import (
    TransformerMixin,
)

from tabpfn.architectures.encoders import (
    MulticlassClassificationTargetEncoderStep,
    TorchPreprocessingPipeline,
)
from tabpfn.constants import (
    REGRESSION_NAN_BORDER_LIMIT_LOWER,
    REGRESSION_NAN_BORDER_LIMIT_UPPER,
)
from tabpfn.preprocessing.datamodel import Feature, FeatureModality, FeatureSchema

if TYPE_CHECKING:
    from sklearn.base import TransformerMixin
    from sklearn.pipeline import Pipeline

    from tabpfn.architectures.interface import Architecture

MAXINT_RANDOM_SEED = int(np.iinfo(np.int32).max)


def get_autocast_context(
    device: torch.device, *, enabled: bool
) -> contextlib.AbstractContextManager:
    """Returns a torch.autocast context manager, disabling it for MPS devices.

    Args:
        device: The torch device being used.
        enabled: Whether to enable autocast.

    Returns:
        A context manager for autocasting.
    """
    if device.type == "mps":
        return contextlib.nullcontext()
    return torch.autocast(device.type, enabled=enabled)


def _repair_borders(borders: np.ndarray, *, inplace: Literal[True]) -> None:
    # Try to repair a broken transformation of the borders:
    #   This is needed when a transformation of the ys leads to very extreme values
    #   in the transformed borders, since the borders spanned a very large range in
    #   the original space.
    #   Borders that were transformed to extreme values are all set to the same
    #   value, the maximum of the transformed borders. Thus probabilities predicted
    #   in these buckets have no effects. The outermost border is set to the
    #   maximum of the transformed borders times 2, so still allow for some weight
    #   in the long tailed distribution and avoid infinite loss.
    if inplace is not True:
        raise NotImplementedError("Only inplace is supported")

    if np.isnan(borders[-1]):
        nans = np.isnan(borders)
        largest = borders[~nans].max()
        borders[nans] = largest
        borders[-1] = borders[-1] * 2

    if borders[-1] - borders[-2] < 1e-6:
        borders[-1] = borders[-1] * 1.1

    if borders[0] == borders[1]:
        borders[0] -= np.abs(borders[0] * 0.1)


def _cancel_nan_borders(
    *,
    borders: np.ndarray,
    broken_mask: npt.NDArray[np.bool_],
) -> tuple[np.ndarray, npt.NDArray[np.bool_]]:
    # OPTIM: You could do one check at a time
    # assert it is consecutive areas starting from both ends
    borders = borders.copy()
    num_right_borders = (broken_mask[:-1] > broken_mask[1:]).sum()
    num_left_borders = (broken_mask[1:] > broken_mask[:-1]).sum()
    assert num_left_borders <= 1
    assert num_right_borders <= 1

    if num_right_borders:
        assert bool(broken_mask[0]) is True
        rightmost_nan_of_left = np.where(broken_mask[:-1] > broken_mask[1:])[0][0] + 1
        borders[:rightmost_nan_of_left] = borders[rightmost_nan_of_left]
        borders[0] = borders[1] - 1.0

    if num_left_borders:
        assert bool(broken_mask[-1]) is True
        leftmost_nan_of_right = np.where(broken_mask[1:] > broken_mask[:-1])[0][0]
        borders[leftmost_nan_of_right + 1 :] = borders[leftmost_nan_of_right]
        borders[-1] = borders[-2] + 1.0

    # logit mask, mask out the nan positions, the borders are 1 more than logits
    logit_cancel_mask = broken_mask[1:] | broken_mask[:-1]
    return borders, logit_cancel_mask


DevicesSpecification = Union[
    torch.device, str, Sequence[Union[torch.device, str]], Literal["auto"]
]


def infer_devices(devices: DevicesSpecification) -> tuple[torch.device, ...]:
    """Selects the appropriate PyTorch devices for inference.

    If `device` is "auto" then the devices are selected as follows:
    1. If CUDA is available and not excluded, returns the first "cuda" device
    2. Otherwise, if MPS is available and not excluded, returns the "mps" device
    3. Otherwise, returns the "cpu" device

    CUDA and MPS can be excluded from the "auto" selection by specifying the
    TABPFN_EXCLUDE_DEVICES environment variable. This can be either "cuda", "mps", or
    "cuda,mps". This is useful for excluding device classes in CI.

    Args:
        devices: The device specification. One of:
            - "auto": the device will be selected as described above
            - a PyTorch device string like "cuda", "mps", or "cpu": this single device
                will be selected by parsing the string to a torch.device
            - a torch.device: this single device will be selected
            - a list of PyTorch device strings or torch.device: each item will be
                converted to a torch.device, and all of the devices selected

    Returns:
        A tuple of at least one device.
    """
    exclude_devices = {
        d.strip()
        for d in os.getenv("TABPFN_EXCLUDE_DEVICES", "").split(",")
        if d.strip()
    }

    if devices == "auto":
        if "cuda" not in exclude_devices and torch.cuda.is_available():
            return (torch.device("cuda:0"),)

        if _is_mps_supported() and "mps" not in exclude_devices:
            return (torch.device("mps"),)

        return (torch.device("cpu"),)

    if isinstance(devices, (str, torch.device)):
        devices = (devices,)

    devices = tuple(_parse_device(device) for device in devices)

    if len(set(devices)) != len(devices):
        raise ValueError(
            "The list of devices for inference cannot contain the same device more "
            f"than once. It contained: {devices}"
        )

    if not _is_mps_supported() and any(d.type == "mps" for d in devices):
        raise ValueError(
            "The MPS device was selected, "
            "but this is not supported by TabPFN before PyTorch 2.5. "
            'Set `device="cpu"` instead.'
        )

    return devices


def _parse_device(device: str | torch.device) -> torch.device:
    # This is safe because torch.device(torch.device(...)) is a noop.
    # torch.device(device) returns a fairly informative error message if `device` is not
    # a valid device, thus do no extra error reporting.
    device = torch.device(device)

    # In older versions of PyTorch, some torch.cuda functions fail if the device has no
    # index. 0 is implicit if no index is specified, so add it.
    if device == torch.device("cuda"):
        device = torch.device("cuda:0")

    return device


def _is_mps_supported() -> bool:
    """Return True if the MPS device is supported, otherwise False.

    We have found that using MPS can lead to poor accuracy on PyTorch <2.5. See
    https://github.com/PriorLabs/TabPFN/pull/619
    """
    return torch.__version__ >= "2.5" and torch.backends.mps.is_available()


def is_autocast_available(device_type: str) -> bool:
    """Infer whether autocast is available for the given device type.

    Args:
        device_type: The device type to check for autocast availability.

    Returns:
        Whether autocast is available for the given device type.
    """
    # Try to use PyTorch's built-in function first
    try:
        # Check if the function is available in torch
        if hasattr(torch.amp.autocast_mode, "is_autocast_available"):
            # Use function directly
            torch_is_autocast_available = torch.amp.autocast_mode.is_autocast_available
            return bool(torch_is_autocast_available(device_type))
        # Fall back to custom implementation
        raise AttributeError("is_autocast_available not found")
    except (ImportError, AttributeError):
        # Fall back to custom implementation if the function isn't available
        return bool(
            hasattr(torch.cuda, "amp")
            and hasattr(torch.cuda.amp, "autocast")
            and (
                device_type == torch.device("cuda").type
                or (
                    device_type == torch.device("cpu").type
                    and hasattr(torch.cpu, "amp")
                )
            ),
        )


def infer_fp16_inference_mode(
    devices: Sequence[torch.device], *, enable: bool | None
) -> bool:
    """Infer whether fp16 inference should be enabled.

    Args:
        devices: The devices to validate against.
        enable:
            Whether it should be enabled, `True` or `False`, otherwise if `None`,
            detect if it's possible and use it if so.

    Returns:
        Whether to use fp16 inference or not.

    Raises:
        ValueError: If fp16 inference was enabled and any of the selected devices do
            not support it.
    """
    is_cpu = any(device.type.lower() == "cpu" for device in devices)
    fp16_available = (
        not is_cpu  # CPU can show enabled, yet it kills inference speed
        and any(is_autocast_available(device.type) for device in devices)
    )

    if enable is None:
        return fp16_available

    if enable is True:
        if not fp16_available:
            raise ValueError(
                "You specified `fp16_inference=True`, however"
                "`torch.amp.autocast_mode.is_autocast_available()`"
                f" reported that one or more of the selected devices ({devices=})"
                " does not support it."
                "\nPlease ensure your version of torch and device type"
                " are compatible with torch.autocast()`"
                " or set `fp16_inference=False`.",
            )
        return True

    if enable is False:
        return False

    raise ValueError(f"Unrecognized argument '{enable}'")


def infer_random_state(
    random_state: int | np.random.RandomState | np.random.Generator | None,
) -> tuple[int, np.random.Generator]:
    """Infer the random state from the given input.

    Args:
        random_state: The random state to infer.

    Returns:
        A static integer seed and a random number generator.
    """
    if isinstance(random_state, (int, np.integer)):
        np_rng = np.random.default_rng(random_state)
        static_seed = int(random_state)
    elif isinstance(random_state, np.random.RandomState):
        static_seed = int(random_state.randint(0, MAXINT_RANDOM_SEED))
        np_rng = np.random.default_rng(static_seed)
    elif isinstance(random_state, np.random.Generator):
        np_rng = random_state
        static_seed = int(np_rng.integers(0, MAXINT_RANDOM_SEED))
    elif random_state is None:
        np_rng = np.random.default_rng()
        static_seed = int(np_rng.integers(0, MAXINT_RANDOM_SEED))
    else:
        raise ValueError(f"Invalid random_state {random_state}")

    return static_seed, np_rng


def _map_to_bucket_ix(y: torch.Tensor, borders: torch.Tensor) -> torch.Tensor:
    ix = torch.searchsorted(sorted_sequence=borders, input=y) - 1
    ix[y == borders[0]] = 0
    ix[y == borders[-1]] = len(borders) - 2
    return ix


# TODO (eddiebergman): Can probably put this back to the Bar distribution.
# However we don't really need the full BarDistribution class and this was
# put here to make that a bit more obvious in terms of what was going on.
def _cdf(logits: torch.Tensor, borders: torch.Tensor, ys: torch.Tensor) -> torch.Tensor:
    ys = ys.repeat((*logits.shape[:-1], 1))
    n_bars = len(borders) - 1
    y_buckets = _map_to_bucket_ix(ys, borders).clamp(0, n_bars - 1).to(logits.device)

    probs = torch.softmax(logits, dim=-1)
    prob_so_far = torch.cumsum(probs, dim=-1) - probs
    prob_left_of_bucket = prob_so_far.gather(index=y_buckets, dim=-1)

    bucket_widths = borders[1:] - borders[:-1]
    share_of_bucket_left = (ys - borders[y_buckets]) / bucket_widths[y_buckets]
    share_of_bucket_left = share_of_bucket_left.clamp(0.0, 1.0)

    prob_in_bucket = probs.gather(index=y_buckets, dim=-1) * share_of_bucket_left
    prob_left_of_ys = prob_left_of_bucket + prob_in_bucket

    prob_left_of_ys[ys <= borders[0]] = 0.0
    prob_left_of_ys[ys >= borders[-1]] = 1.0
    return prob_left_of_ys.clip(0.0, 1.0)


def translate_probs_across_borders(
    logits: torch.Tensor,
    *,
    frm: torch.Tensor,
    to: torch.Tensor,
) -> torch.Tensor:
    """Translate the probabilities across the borders.

    Args:
        logits: The logits defining the distribution to translate.
        frm: The borders to translate from.
        to: The borders to translate to.

    Returns:
        The translated probabilities.
    """
    prob_left = _cdf(logits, borders=frm, ys=to)
    prob_left[..., 0] = 0.0
    prob_left[..., -1] = 1.0

    return (prob_left[..., 1:] - prob_left[..., :-1]).clamp_min(0.0)


def remove_non_differentiable_preprocessing_from_models(
    models: list[Architecture],
) -> None:
    """Remove non-differentiable encoder steps from the model.

    Args:
        models: The models to update.
    """
    for model in models:
        diffable_steps = []  # only differentiable encoder steps.
        for module in model.y_encoder:
            if isinstance(module, MulticlassClassificationTargetEncoderStep):
                pass
            else:
                diffable_steps.append(module)

        model.y_encoder = TorchPreprocessingPipeline(
            steps=diffable_steps, output_key="output"
        )


def transform_borders_one(
    borders: np.ndarray,
    target_transform: TransformerMixin | Pipeline,
    *,
    repair_nan_borders_after_transform: bool,
) -> tuple[npt.NDArray[np.bool_] | None, bool, np.ndarray]:
    """Transforms the borders used for the bar distribution for regression.

    Args:
        borders: The borders to transform.
        target_transform: The target transformer to use.
        repair_nan_borders_after_transform:
            Whether to repair any borders that are NaN after the transformation.

    Returns:
        logit_cancel_mask:
            The mask of the logit values to ignore,
            those that mapped to NaN borders.
        descending_borders: Whether the borders are descending after transformation
        borders_t: The transformed borders themselves.
    """
    borders_t = target_transform.inverse_transform(borders.reshape(-1, 1)).squeeze()  # type: ignore

    logit_cancel_mask: npt.NDArray[np.bool_] | None = None
    if repair_nan_borders_after_transform:
        broken_mask = (
            ~np.isfinite(borders_t)
            | (borders_t > REGRESSION_NAN_BORDER_LIMIT_UPPER)
            | (borders_t < REGRESSION_NAN_BORDER_LIMIT_LOWER)
        )
        if broken_mask.any():
            borders_t, logit_cancel_mask = _cancel_nan_borders(
                borders=borders_t,
                broken_mask=broken_mask,
            )

    _repair_borders(borders_t, inplace=True)

    reversed_order = np.arange(len(borders_t) - 1, -1, -1)
    descending_borders = (np.argsort(borders_t) == reversed_order).all()
    if descending_borders:
        borders_t = borders_t[::-1]
        logit_cancel_mask = (
            logit_cancel_mask[::-1] if logit_cancel_mask is not None else None
        )

    return logit_cancel_mask, descending_borders, borders_t


def pad_tensors(
    tensor_list: list[torch.Tensor],
    padding_val: float | None = 0,
    *,
    labels: bool = False,
) -> list[torch.Tensor]:
    """Pad tensors to maximum dims at the last dimensions.
    if labels=False, 2d tensors are expected, if labels=True, one 1d
    vectors are expected as inputs.

    Args:
        tensor_list: List of tensors to be padded.
        padding_val: what value to use for padding.
        labels: If true, the tensor list should contain 1D
            tensors that are padded only along this dimension.
            If false, rows and feature dimensions are padded.
    """
    max_size_clms = max([item.size(-1) for item in tensor_list])
    if not labels:
        max_size_rows = max([item.size(-2) for item in tensor_list])
    ret_list = []
    for item in tensor_list:
        pad_seqence = [0, max_size_clms - item.size(-1)]
        if not labels:
            pad_seqence.extend([0, max_size_rows - item.size(-2)])
        padded_item = torch.nn.functional.pad(
            item, pad_seqence, mode="constant", value=padding_val
        )
        ret_list.append(padded_item)
    return ret_list


def balance_probas_by_class_counts(
    probas: torch.Tensor,
    class_counts: np.ndarray,
) -> torch.Tensor:
    """Balance probabilities by class counts.

    Args:
        probas: The probabilities to balance.
        class_counts: The class counts to use for balancing.

    Returns:
        The balanced probabilities.
    """
    class_prob_in_train = class_counts / class_counts.sum()
    balanced_probas = probas / torch.from_numpy(class_prob_in_train).float().to(
        probas.device
    )
    return balanced_probas / balanced_probas.sum(dim=-1, keepdim=True)


def convert_batch_of_cat_ix_to_schema(
    batch_of_cat_indices: list[list[list[int]]],
    num_features: int,
) -> list[list[FeatureSchema]]:
    """Convert a batch of categorical indices to a schema."""
    feature_schema = []
    for ibatch in batch_of_cat_indices:
        feature_schema.append([])
        for cat_indices in ibatch:
            features = [
                Feature(
                    name=f"c{i}",
                    modality=FeatureModality.CATEGORICAL
                    if i in cat_indices
                    else FeatureModality.NUMERICAL,
                )
                for i in range(num_features)
            ]
            feature_schema[-1].append(FeatureSchema(features=features))

    return feature_schema



"""Differentiable Z-Norm Step."""

from __future__ import annotations

from typing import TYPE_CHECKING
from typing_extensions import override

import torch

from tabpfn.preprocessing.pipeline_interface import (
    PreprocessingStep,
)

if TYPE_CHECKING:
    from tabpfn.preprocessing.datamodel import FeatureSchema


class DifferentiableZNormStep(PreprocessingStep):
    """Differentiable Z-Norm Step."""

    def __init__(self):
        super().__init__()

        self.means = torch.tensor([])
        self.stds = torch.tensor([])

    @override
    def _fit(  # type: ignore
        self,
        X: torch.Tensor,
        feature_schema: FeatureSchema,
    ) -> FeatureSchema:
        self.means = X.mean(dim=0, keepdim=True)
        self.stds = X.std(dim=0, keepdim=True)
        return feature_schema

    @override
    def _transform(  # type: ignore
        self, X: torch.Tensor, *, is_test: bool = False
    ) -> tuple[torch.Tensor, None, None]:
        assert X.shape[1] == self.means.shape[1]
        assert X.shape[1] == self.stds.shape[1]
        return (X - self.means) / self.stds, None, None


__all__ = [
    "DifferentiableZNormStep",
]



"""DEPRECATION WARNING: Please note that this module will be deprecated in future
versions of TabPFN in favor of a new torch-based preprocessing pipeline.
"""

from .pipeline_interfaces import (
    TorchPreprocessingPipeline,
    TorchPreprocessingStep,
)
from .steps import (
    FeatureTransformEncoderStep,
    FrequencyFeatureEncoderStep,
    LinearInputEncoderStep,
    MLPInputEncoderStep,
    MulticlassClassificationTargetEncoderStep,
    NanHandlingEncoderStep,
    NormalizeFeatureGroupsEncoderStep,
    RemoveDuplicateFeaturesEncoderStep,
    RemoveEmptyFeaturesEncoderStep,
)

__all__ = (
    "FeatureTransformEncoderStep",
    "FrequencyFeatureEncoderStep",
    "LinearInputEncoderStep",
    "MLPInputEncoderStep",
    "MulticlassClassificationTargetEncoderStep",
    "NanHandlingEncoderStep",
    "NormalizeFeatureGroupsEncoderStep",
    "RemoveDuplicateFeaturesEncoderStep",
    "RemoveEmptyFeaturesEncoderStep",
    "TorchPreprocessingPipeline",
    "TorchPreprocessingStep",
)
